# coding=utf-8
"""
Folders / Documents module
"""
from __future__ import absolute_import

from abilian.sbe.extension import sbe


def register_plugin(app):
  sbe.init_app(app)
  from .views import documents
  from .models import setup_listener
  from .commands import manager
  from . import events
  from . import lock

  app.register_blueprint(documents)
  setup_listener()

  # set default lock lifetime
  app.config.setdefault('SBE_LOCK_LIFETIME', lock.DEFAULT_LIFETIME)

  if app.script_manager:
    app.script_manager.add_command('documents', manager)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import logging
from datetime import datetime
import time
from os.path import dirname, join

from blinker import ANY
import sqlalchemy as sa
from sqlalchemy import (
  Column, Unicode, ForeignKey, Boolean, DateTime,
  Integer, UniqueConstraint, String, and_,
)
from sqlalchemy.orm import relation, relationship, backref
from sqlalchemy.orm.attributes import OP_APPEND, OP_REMOVE
from flask import current_app

from abilian.i18n import _l
from abilian.core.extensions import db
from abilian.core.models import NOT_AUDITABLE, SEARCHABLE
from abilian.core.models.subjects import Group, User
from abilian.core.models.blob import Blob
from abilian.core.entities import Entity
from abilian.services.indexing import indexable_role
from abilian.services.security import (
  Role, security, RoleType, Admin,
  Reader as READER,
  Writer as WRITER,
  Manager as MANAGER,
)

from abilian.sbe.apps.documents.repository import repository
from abilian.sbe.apps.documents.models import Folder
from . import signals

logger = logging.getLogger(__name__)

MEMBER = Role('member', label=_l(u'role_member'), assignable=False)
VALID_ROLES = frozenset((READER, WRITER, MANAGER, MEMBER,))


class Membership(db.Model):
  """
  Objects that represent the membership of someone in a community.
  """
  __tablename__ = 'community_membership'

  id = Column(Integer, primary_key=True)

  user_id = Column(ForeignKey('user.id'), index=True, nullable=False)
  user = relationship(User, lazy='joined',
                      backref=backref('communautes_membership', lazy='select'))

  community_id = Column(ForeignKey('community.id'), index=True, nullable=False)
  community = relationship('Community', lazy='joined')

  role = Column(RoleType())  # should be either 'member' or 'manager'

  __table_args__ = (UniqueConstraint('user_id', 'community_id'),)

  def __repr__(self):
    return u'<Membership user={}, community={}, role={}>'.format(
      repr(self.user),
      repr(self.community),
      str(self.role)
    ).encode('utf-8')


def community_content(cls):
  """
  Class decorator to mark models considered as community content. This is
  required for proper indexation.
  """
  cls.is_community_content = True

  def community_slug(self):
    return self.community and self.community.slug

  cls.community_slug = property(community_slug)
  index_to = cls.__indexation_args__.setdefault('index_to', ())
  index_to += (
    ('community_slug', ('community_slug',),),
  )
  cls.__indexation_args__['index_to'] = index_to

  def _indexable_roles_and_users(self):
    if not self.community:
      return []
    return indexable_roles_and_users(self.community)

  cls._indexable_roles_and_users = property(_indexable_roles_and_users)
  return cls


def indexable_roles_and_users(community):
  """
  Mixin to use to replace Entity._indexable_roles_and_users.

  Will be removed when communities are upgraded to use standard role based
  access (by setting permissions and using security service).
  """
  return u' '.join(indexable_role(user)
                   for user in community.members)


class Community(Entity):
  """
  Ad-hoc objects that hold properties about a community.
  """
  __indexation_args__ = {}
  __indexation_args__.update(Entity.__indexation_args__)
  index_to = __indexation_args__.setdefault('index_to', ())
  index_to += (('id', ('id', 'community_id')),)
  __indexation_args__['index_to'] = index_to
  del index_to

  is_community_content = True

  # : A public description.
  description = Column(Unicode(500), default=u"", nullable=False,
                       info=SEARCHABLE)

  #: An image or logo for this community.
  image_id = Column(ForeignKey(Blob.id), index=True)
  image = relationship(Blob, lazy='joined')

  #: The root folder for this community.
  folder_id = Column(Integer,
                     ForeignKey(Folder.id, use_alter=True,
                                name='fk_community_root_folder'),
                     unique=True)
  folder = relation(Folder,
                    single_parent=True,  # required for delete-orphan
                    primaryjoin=(folder_id == Folder.id),
                    cascade='all, delete-orphan',
                    backref=backref('_community', lazy='select', uselist=False))

  #: The group this community is linked to, if any. Memberships are then
  #: reflected
  group_id = Column(ForeignKey(Group.id), nullable=True, unique=True)
  group = relation(Group, foreign_keys=group_id, lazy='select')

  #: Memberships for this community.
  memberships = relationship(Membership, cascade="all, delete-orphan")

  #: direct access to :class:`User` members
  members = relationship(
    User,
    secondary=Membership.__table__,
    viewonly=True,
    backref=backref('communities', lazy='select', viewonly=True),
  )

  #: Number of members in this community.
  membership_count = Column(Integer, default=0, nullable=False,
                            info=NOT_AUDITABLE)

  #: Number of documents in this community.
  document_count = Column(Integer, default=0, nullable=False, info=NOT_AUDITABLE)

  #: Last time something happened in this community
  last_active_at = Column(DateTime, default=datetime.utcnow, nullable=False,
                          info=NOT_AUDITABLE)

  # Various features for this community:

  #: True if this community has a document management space
  has_documents = Column(Boolean, nullable=False, default=True)

  #: True if this community has a wiki
  has_wiki = Column(Boolean, nullable=False, default=True)

  #: True if this community has a forum
  has_forum = Column(Boolean, nullable=False, default=True)

  #: One of 'participative' or 'informative
  type = Column(String(20), nullable=False, default="informative")

  #: One of 'secret', 'public', 'open' (currently not used)
  visibility = Column(String(20), nullable=False, default="secret")

  #: Used for segmenting communities. Not used currently.
  category = Column(String(20), nullable=False, default="")

  #: True if regular members can send stuff by email to all members
  # Not used anymore.
  members_can_send_by_email = Column(Boolean, nullable=False, default=False)

  def __init__(self, **kw):
    self.has_documents = True
    self.membership_count = 0
    self.document_count = 0
    self.members_can_send_by_email = False
    Entity.__init__(self, **kw)
    if self.has_documents and not self.folder:
      #FIXME: this should be done in documents by using signals
      name = self.name
      if not name:
        # during creation, we may have to provide a temporary name for
        # subfolder, we don't want empty names on folders since they must be
        # unique among siblings
        name = u'{}_{}-{}'.format(self.__class__.__name__, str(self.id),
                                  time.asctime())
      self.folder = repository.root_folder.create_subfolder(name)
      #if not self.group:
    #  self.group = Group(name=self.name)

    if not self.image:
      fn = join(dirname(__file__), "data", "community.png")
      self.image = Blob(open(fn).read())

  def rename(self, name):
    self.name = name
    if self.folder:
      #FIXME: use signals
      self.folder.name = name

  def get_memberships(self, role=None):
    M = Membership
    memberships = M.query.filter(M.community_id == self.id)
    if role:
      memberships = memberships.filter(M.role == role)
    return memberships.all()

  def set_membership(self, user, role):
    """
    Add a member with the given role, or set the role of an existing
    member.
    """
    assert isinstance(user, User)
    role = Role(role)

    if role not in VALID_ROLES:
      raise ValueError("Invalid role: {}".format(role))

    session = sa.orm.object_session(self) or db.session()
    is_new = True
    M = Membership
    membership = session.query(M) \
                        .filter(and_(M.user_id == user.id,
                                     M.community_id == self.id))\
                        .first()

    if not membership:
      membership = Membership(community=self, user=user, role=role)
      session.add(membership)
      self.membership_count += 1
    else:
      is_new = False
      membership.role = role

    signals.membership_set.send(self, membership=membership, is_new=is_new)

  def remove_membership(self, user):
    M = Membership
    membership = M.query \
      .filter(and_(M.user_id == user.id, M.community_id == self.id)).first()
    if not membership:
      raise KeyError(
        "User {} is not a member of community {}".format(user, self))

    db.session.delete(membership)
    self.membership_count -= 1
    signals.membership_removed.send(self, membership=membership)

  def update_roles_on_folder(self):
    if self.folder:
      self.ungrant_all_roles_on_folder()
      for membership in self.memberships:
        user = membership.user
        role = membership.role
        if role == MANAGER:
          security.grant_role(user, MANAGER, self.folder)
        else:
          if self.type == 'participative':
            security.grant_role(user, WRITER, self.folder)
          else:
            security.grant_role(user, READER, self.folder)

  def ungrant_all_roles_on_folder(self):
    if self.folder:
      role_assignments = security.get_role_assignements(self.folder)
      for principal, role in role_assignments:
        security.ungrant_role(principal, role, self.folder)

  def get_role(self, user):
    """
    Returns the current user's role in this community.
    """
    M = Membership
    membership = current_app.db.session()\
        .query(M.role)\
        .filter(and_(M.community_id == self.id,
                     M.user_id == user.id))\
        .first()

    return membership.role if membership else None

  def has_member(self, user):
    return self.get_role(user) is not None

  def has_permission(self, user, permission):
    if user.has_role(Admin):
      return True
    role = self.get_role(user)
    if role == MANAGER:
      return True
    if role == MEMBER and permission in ('read', 'write'):
      return True
    return False

  def touch(self):
    self.last_active_at = datetime.utcnow()

  @property
  def _indexable_roles_and_users(self):
    return indexable_roles_and_users(self)


def CommunityIdColumn():
  return Column(
    ForeignKey(Community.id),
    nullable=False,
    info=SEARCHABLE | dict(
      index_to=(
        ('community_id', ('community_id',)),
      )),
  )


# Handlers to keep community/group members in sync
#
_PROCESSED_ATTR = '__sbe_community_group_sync_processed__'

@signals.membership_set.connect_via(ANY)
def _membership_added(sender, membership, is_new):
  if not is_new:
    return

  if getattr(membership.user, _PROCESSED_ATTR, False) is OP_APPEND:
    return

  if sender.group and membership.user not in sender.group.members:
    logger.debug('_membership_added(%r, %r, %r) user: %r',
                 sender, membership, is_new, membership.user)
    setattr(membership.user, _PROCESSED_ATTR, OP_APPEND)
    sender.group.members.add(membership.user)


@signals.membership_removed.connect_via(ANY)
def membership_removed(sender, membership):
  if getattr(membership.user, _PROCESSED_ATTR, False) is OP_REMOVE:
    return

  if sender.group and membership.user in sender.group.members:
    logger.debug('_membership_removed(%r, %r) user: %r',
                 sender, membership, membership.user)
    setattr(membership.user, _PROCESSED_ATTR, OP_REMOVE)
    sender.group.members.discard(membership.user)


@sa.event.listens_for(Community.members, 'append')
@sa.event.listens_for(Community.members, 'remove')
def _on_member_change(community, user, initiator):
  group = community.group
  if not group:
    return

  logger.debug('_on_member_change(%r, %r, op=%r)',
               community, user, initiator.op)

  if getattr(user, _PROCESSED_ATTR, False) is initiator.op:
    return

  setattr(user, _PROCESSED_ATTR, initiator.op)

  if initiator.op is OP_APPEND:
    if user not in group.members:
      group.members.add(user)

  elif initiator.op is OP_REMOVE:
    if user in group.members:
      group.members.discard(user)


@sa.event.listens_for(Community.group, 'set', active_history=True)
def _on_linked_group_change(community, value, oldvalue, initiator):
  if value == oldvalue:
    return

  logger.debug('_on_linked_group_change(%r, %r, %r)',
               community, value, oldvalue)

  if oldvalue is not None and oldvalue.members:
    logger.debug('_on_linked_group_change(%r, %r, %r): oldvalue clear()',
                 community, value, oldvalue)
    oldvalue.members.clear()

  members = set(community.members)
  if value is not None and value.members != members:
    logger.debug('_on_linked_group_change(%r, %r, %r): set value.members',
                 community, value, oldvalue)
    value.members = members


def _safe_get_community(group):
  session = sa.orm.object_session(group)
  if not session:
    return None

  with session.no_autoflush:
    try:
      return session.query(Community)\
                    .filter(Community.group == group)\
                    .options(sa.orm.joinedload(Community.group),
                             sa.orm.joinedload(Community.members))\
                    .one()
    except sa.orm.exc.NoResultFound:
      return None


@sa.event.listens_for(Group.members, 'append')
@sa.event.listens_for(Group.members, 'remove')
def _on_group_member_change(group, user, initiator):
  community = _safe_get_community(group)

  if not community:
    return

  op = initiator.op
  if getattr(user, _PROCESSED_ATTR, False) is op:
    return

  is_present = user in community.members
  setattr(user, _PROCESSED_ATTR, op)
  logger.debug('_on_group_member_change(%r, %r, op=%r) community: %r',
               group, user, initiator.op, community)

  if ((op is OP_APPEND and is_present)
      or (op is OP_REMOVE and not is_present)):
    return

  if op is OP_APPEND:
    community.set_membership(user, MEMBER)

  elif op is OP_REMOVE:
    community.remove_membership(user)



@sa.event.listens_for(Group.members, 'set', active_history=True)
def _on_group_members_replace(group, value, oldvalue, initiator):
  if value == oldvalue:
    return

  community = _safe_get_community(group)
  if not community:
    return

  members = set(community.members)
  logger.debug('_on_group_members_replace(%r, %r, %r) community: %r',
               group, value,  oldvalue, community)

  for u in members - value:
    if getattr(u, _PROCESSED_ATTR, False) is OP_REMOVE:
      continue
    setattr(u, _PROCESSED_ATTR, OP_REMOVE)
    community.remove_membership(u)

  for u in value - members:
    if getattr(u, _PROCESSED_ATTR, False) is OP_APPEND:
      continue
    setattr(u, _PROCESSED_ATTR, OP_APPEND)
    community.set_membership(u, MEMBER)



# coding=utf-8
"""
Celery tasks related to document transformation and preview.
"""
from __future__ import absolute_import

import logging
from contextlib import contextmanager

from celery import shared_task
from guess_language import guessLanguageName

from abilian.services.conversion import ConversionError
from abilian.services import get_service, converter
from abilian.core.extensions import db


logger = logging.getLogger(__package__)

@contextmanager
def get_document(document_id, session=None):
  """
  Context manager that yields (session, document)
  """
  from .models import Document
  doc_session = session
  if session is None:
    doc_session = db.create_scoped_session()

  with doc_session.begin_nested():
    query = doc_session.query(Document)
    document = query.get(document_id)
    yield (doc_session, document)

  # cleanup
  if session is None:
    doc_session.commit()
    doc_session.close()


@shared_task
def process_document(document_id):
  """
  Run document processing chain
  """
  with get_document(document_id) as (session, document):
    if document is None:
      return

    # True = Ok, None means no check performed (no antivirus present)
    is_clean = _run_antivirus(document)
    if is_clean is False:
      return

  preview_document.delay(document_id,)
  convert_document_content.delay(document_id,)


def _run_antivirus(document):
  antivirus = get_service('antivirus')
  if antivirus and antivirus.running:
    is_clean = antivirus.scan(document.content_blob)
    if 'antivirus_task' in document.content_blob.meta:
      del document.content_blob.meta['antivirus_task']
    return is_clean
  return None


@shared_task
def antivirus_scan(document_id):
  """
  Return antivirus.scan() result
  """
  with get_document(document_id) as (session, document):
    if document is None:
      return
    return _run_antivirus(document)


@shared_task
def preview_document(document_id):
  """Computes the document preview images with its default preview size.
  """
  with get_document(document_id) as (session, document):
    if document is None:
      # deleted after task queued, but before task run
      return

    try:
      converter.to_image(document.content_digest, document.content,
                         document.content_type, 0, document.preview_size)
    except ConversionError as e:
      logger.info('Preview failed: %s', str(e),
                  exc_info=True, extra={'stack': True})


@shared_task
def convert_document_content(document_id):
  """Converts document content.
  """
  with get_document(document_id) as (session, document):
    if document is None:
      # deleted after task queued, but before task run
      return

    error_kwargs = dict(exc_info=True, extra={'stack': True})

    conversion_args = (document.content_digest, document.content,
                       document.content_type)

    if document.content_type == "application/pdf":
      document.pdf = document.content
    else:
      try:
        document.pdf = converter.to_pdf(*conversion_args)
      except ConversionError as e:
        document.pdf = ""
        logger.info("Conversion to PDF failed: %s", str(e), **error_kwargs)

    try:
      document.text = converter.to_text(document.content_digest, document.content,
                                        document.content_type)
    except ConversionError as e:
      document.text = u""
      logger.info("Conversion to text failed: %s", str(e), **error_kwargs)

    document.extra_metadata = {}
    try:
      document.extra_metadata = converter.get_metadata(*conversion_args)
    except ConversionError as e:
      logger.warning("Metadata extraction failed: %s", str(e),
                     **error_kwargs)
    except UnicodeDecodeError as e:
      logger.error("Unicode issue: %s", str(e), **error_kwargs)
    except Exception as e:
      logger.error("Other issue: %s", str(e), **error_kwargs)

    if document.text:
      document.language = guessLanguageName(document.text)

    document.page_num = document.extra_metadata.get("PDF:Pages", 1)



# coding=utf-8
""" Communities module
"""
from __future__ import absolute_import

from abilian.sbe.extension import sbe


def register_plugin(app):
  sbe.init_app(app)

  # Used for side-effect
  from . import events  # noqa

  from .views import communities
  app.register_blueprint(communities)

  from . import search
  search.init_app(app)



#noinspection PyMissingConstructor
class Env(dict):
  """
  A dictionnary whose objects can be accessed by '.' notation, saving a few
  keystrokes.
  """

  def __init__(self, label=None, **kw):
    #self.bread_crumbs = self.breadcrumbs = make_bread_crumbs(label=label)
    for key, value in kw.items():
      self[key] = value

  def __setattr__(self, key, value):
    self[key] = value

  def __getattr__(self, key):
    return self[key]



"""
Social content items: messages aka status updates, private messages, etc.
"""
import re

from sqlalchemy.orm.query import Query
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.types import Integer, UnicodeText, LargeBinary

from abilian.core.models.subjects import User, Group
from abilian.core.entities import Entity, SEARCHABLE
from abilian.core.extensions import db

__all__ = ['Message', 'PrivateMessage']


class MessageQuery(Query):

  def by_creator(self, user):
    return self.filter(Message.creator_id == user.id)


class Message(Entity):
  """Message aka Status update aka Note.

  See: http://activitystrea.ms/head/activity-schema.html#note
  """
  __tablename__ = 'message'
  __indexable__ = False
  __editable__ = ['content']
  __exportable__ = __editable__ + ['id', 'created_at', 'updated_at', 'creator_id', 'owner_id']

  #: The content for this message.
  content = Column(UnicodeText(), info=SEARCHABLE|dict(index_to=('text',)))

  #: Nullable: if null, then message is public.
  group_id = Column(Integer, ForeignKey(Group.id))

  #: The group this message has been posted to.
  group = relationship('Group', primaryjoin=(group_id == Group.id), lazy='joined')

  query = db.session.query_property(MessageQuery)

  @property
  def tags(self):
    return re.findall("#([^\W]+)(?u)", self.content)


# TODO: inheriting from Entity is overkill here
class TagApplication(Entity):
  __tablename__ = 'tag_application'

  tag = Column(UnicodeText, index=True)
  message_id = Column(Integer, ForeignKey('message.id'))


class PrivateMessage(Entity):
  """Private messages are like messages, except they are private.
  """

  __tablename__ = 'private_message'
  __indexable__ = False
  __editable__ = ['content', 'recipient_id']
  __exportable__ = __editable__ + ['id', 'created_at', 'updated_at', 'creator_id', 'owner_id']

  content = Column(UnicodeText, info=SEARCHABLE|dict(index_to=('text',)))
  recipient_id = Column(Integer, ForeignKey(User.id), nullable=False)


class Like(Entity):

  __tablename__ = 'like'
  __indexable__ = False
  __editable__ = ['content', 'message_id']
  __exportable__ = __editable__ + ['id', 'created_at', 'updated_at', 'creator_id', 'owner_id']

  content = Column(UnicodeText, info=SEARCHABLE|dict(index_to=('text',)))
  message_id = Column(Integer, ForeignKey(Message.id), nullable=False)



from abilian import testing

from .app import Application


class TestConfig(testing.TestConfig):
  CELERY_ALWAYS_EAGER = True # run tasks locally, no async
  CELERY_EAGER_PROPAGATES_EXCEPTIONS = True


class BaseTestCase(testing.BaseTestCase):
  application_class = Application
  config_class = TestConfig

  def get_setup_config(self):
    """ Called before creating application class
    """
    config = testing.BaseTestCase.get_setup_config(self)
    if hasattr(self, 'no_login'):
      config.NO_LOGIN = self.no_login

    return config



# coding=utf-8
"""
"""
from __future__ import absolute_import

from datetime import datetime

from flask import g
import sqlalchemy as sa
from sqlalchemy import (
  UnicodeText, Unicode, Column, ForeignKey, Integer,
  UniqueConstraint, DateTime,
)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.hybrid import hybrid_property

from abilian.core.entities import Entity, db, SEARCHABLE
from abilian.core.models.subjects import User

from abilian.sbe.apps.communities.models import (
  Community, CommunityIdColumn, community_content,
)
from abilian.sbe.apps.documents.models import BaseContent

from . import markup


__all__ = ['WikiPage']


@community_content
class WikiPage(Entity):
  __tablename__ = 'wiki_page'

  # : The title for this page
  _title = Column('title', Unicode(200), nullable=False, index=True)

  community_id = CommunityIdColumn()
  #: The community this page belongs to
  community = relationship(Community,
                           primaryjoin=(community_id == Community.id),
                           backref=backref("wiki",
                                           cascade="all, delete-orphan"))

  #: The body, using some markup language (Markdown for now)
  body_src = Column(UnicodeText, default=u"", nullable=False,
                    info=SEARCHABLE | dict(index_to=('text',)))

  __table_args__ = (UniqueConstraint('title', 'community_id'),)

  def __init__(self, title=u"", body_src=u"", message=u"", *args, **kwargs):
    Entity.__init__(self, *args, **kwargs)
    self.title = title
    self.create_revision(body_src, message)

  # title is defined has an hybrid property to allow name <-> title sync (2 way)
  @hybrid_property
  def title(self):
    return self._title

  @title.setter
  def title(self, title):
    # set title before setting name, so that we don't enter an infinite loop
    # with _wiki_sync_name_title
    self._title = title
    if self.name != title:
      self.name = title

  def create_revision(self, body_src, message=u""):
    revision = WikiPageRevision()
    if self.revisions:
      revision.number = max(r.number for r in self.revisions) + 1
    else:
      revision.number = 0
    self.body_src = revision.body_src = body_src
    revision.message = message
    revision.author = g.user
    revision.page = self

  @property
  def last_revision(self):
    return WikiPageRevision.query\
        .filter(WikiPageRevision.page == self)\
        .order_by(WikiPageRevision.number.desc())\
        .first()

  @property
  def body_html(self):
    html = markup.convert(self, self.body_src)
    return html

    # TODO: remove Javascript from content
    #return bleach.clean(html, strip=True)


@sa.event.listens_for(WikiPage.name, "set", active_history=True)
def _wiki_sync_name_title(entity, new_value, old_value, initiator):
  """
  Synchronize wikipage name -> title.
  wikipage.title -> name is done via hybrid_property, avoiding infinite
  loop (since "set" is received before attribute has received value)
  """
  if entity.title != new_value:
    entity.title = new_value
  return new_value


class WikiPageRevision(db.Model):
  __tablename__ = 'wiki_page_revision'
  __table_args__ = (UniqueConstraint('page_id', 'number'),)

  # : Some primary key just in case
  id = Column(Integer, primary_key=True)

  #: Date and time this revision was created
  created_at = Column(DateTime, default=datetime.utcnow)

  #: The page this revision belongs to
  page = relationship(WikiPage, backref='revisions')
  page_id = Column(ForeignKey(WikiPage.id))

  #: The revision number
  number = Column(Integer, nullable=False)

  #: The body, using some markup language (Markdown for now)
  body_src = Column(UnicodeText, default=u"", nullable=False)

  #: Commit message
  message = Column(UnicodeText, default=u"", nullable=False)

  #: The author of this revision
  author = relationship(User)
  author_id = Column(ForeignKey(User.id))


class WikiPageAttachment(BaseContent):
  __tablename__ = None
  __indexable__ = False
  __mapper_args__ = {'polymorphic_identity': 'wikipage_attachment'}
  sbe_type = 'wikipage:attachment'

  _wikipage_id = Column(Integer, ForeignKey(WikiPage.id), nullable=True)
  wikipage = relationship(WikiPage,
                          primaryjoin=(_wikipage_id == WikiPage.id),
                          backref=backref('attachments', lazy='select',
                                          order_by='WikiPageAttachment.name',
                                          cascade='all, delete-orphan'))



"""
NOTE: this code is a legacy from the early days of the application,
and currently not used.
"""

import json

from flask import request, Blueprint, g, make_response
from flask_login import login_required
from abilian.core import signals
from abilian.core.util import get_params
from abilian.core.models.subjects import User, Group
from abilian.core.extensions import db

from .models import Message


__all__ = ['restapi']

restapi = Blueprint('restapi', __name__, url_prefix="/api")


# Util

def make_json_response(obj, response_code=200):
  if type(obj) == list:
    obj = [ x.to_dict() if hasattr(x, 'to_dict') else x for x in obj ]

  if hasattr(obj, 'to_json'):
    response = make_response(obj.to_json(), response_code)
  elif hasattr(obj, 'to_dict'):
    response = make_response(json.dumps(obj.to_dict()), response_code)
  else:
    response = make_response(json.dumps(obj), response_code)
  response.mimetype = 'application/json'
  return response


#
# Users
#

#[POST] /api/users/USER_ID	Create User Profile
@restapi.route("/users", methods=['POST'])
@login_required
def create_user():
  d = get_params(User.__editable__)
  user = User(**d)
  db.session.add(user)
  db.session.commit()
  return make_json_response(user, 201)


#[GET] /api/users	List Users in Your Organization
@restapi.route("/users")
@login_required
def list_users():
  #l = [ u.to_dict() for u in User.query.all() ]
  l = list(User.query.all())
  return make_json_response(l)


#[GET] /api/users/USER_ID	View User Profile
@restapi.route("/users/<int:user_id>")
@login_required
def get_user(user_id):
  user = User.query.get(user_id)
  return make_json_response(user)


#[GET] /api/users/USER_ID/messages	View Stream of Messages by User
@restapi.route("/users/<int:user_id>/messages")
@login_required
def user_stream(user_id):
  user = User.query.get(user_id)
  messages = Message.query.by_creator(user).all()
  #messages = list(user.messages)
  return make_json_response(messages)


#[PUT] /api/users/USER_ID	Update User Profile
@restapi.route("/users/<int:user_id>", methods=['PUT'])
@login_required
def update_user(user_id):
  user = User.query.get(user_id)
  d = get_params(User.__editable__)
  user.update(d)
  db.session.commit()
  return make_json_response(user)


#[DELETE] /api/users/USER_ID	Deactivate a User
@restapi.route("/users/<int:user_id>", methods=['DELETE'])
@login_required
def delete_user(user_id):
  user = User.query.get(user_id)
  db.session.delete(user)
  db.session.commit()
  return make_response("", 204)


#
# Social graph: following
#

#[GET] /api/users/USER_ID/followers	View Followers of User
@restapi.route("/users/<int:user_id>/followers")
@login_required
def get_followers(user_id):
  user = User.query.get(user_id)
  followers = list(user.followers)
  return make_json_response(followers)


#[GET] /api/users/USER_ID/followees	View List of Users Being Followed
@restapi.route("/users/<int:user_id>/followees")
@login_required
def get_followees(user_id):
  user = User.query.get(user_id)
  followees = list(user.followees)
  return make_json_response(followees)


#[POST] /api/users/USER_ID/followers	Follow a User
@restapi.route("/users/<int:user_id>/followers", methods=['POST'])
@login_required
def follow(user_id):
  user = User.query.get(user_id)
  g.user.follow(user)
  db.session.commit()
  return make_json_response("", 204)


#[DELETE] /api/users/USER_ID/followers/CONTACT_USER_ID	Unfollow a User
@restapi.route("/users/<int:user_id>/followers/<int:contact_user_id>", methods=['DELETE'])
@login_required
def unfollow(user_id, contact_user_id):
  user = User.query.get(user_id)
  g.user.unfollow(user)
  db.session.commit()
  return make_json_response("", 204)


#
# Social graph: groups
#

# [GET] /api/groups	Listing All Groups
@restapi.route("/groups")
@login_required
def list_groups():
  l = list(Group.query.all())
  return make_json_response(l)


# [GET] /api/groups/GROUP_ID	Show a Single Group
@restapi.route("/groups/<int:group_id>")
@login_required
def get_group(group_id):
  group = Group.query.get(group_id)
  return make_json_response(group)


# [GET] /api/groups/GROUP_ID/members	Listing Members of a Group
@restapi.route("/groups/<int:group_id>/members")
@login_required
def get_group_members(group_id):
  group = Group.query.get(group_id)
  return make_json_response(group.members)

# [GET] /api/group_memberships	Listing Group Memberships


# [POST] /api/groups	Create a Group
@restapi.route("/groups", methods=['POST'])
@login_required
def create_group():
  d = get_params(Group.__editable__)
  group = Group(**d)
  db.session.add(group)
  db.session.commit()
  return make_json_response(group, 201)


# [PUT] /api/groups/GROUP_ID	Updating Existing Group

# [DELETE] /api/groups/GROUP_ID/archive	Archiving a Group

# [DELETE] /api/groups/GROUP_ID	Destroy an Archived Message


#
# Messages
#

#[POST] /api/messages	Creating New Messages
@restapi.route("/messages", methods=['POST'])
@login_required
def create_message():
  d = get_params(Message.__editable__)
  message = Message(creator_id=g.user.id, **d)
  db.session.add(message)
  db.session.commit()
  return make_json_response(message, 201)


#[GET] /api/messages	Reading Stream Messages
@restapi.route("/messages")
@login_required
def get_messages():
  messages = list(Message.query.all())
  return make_json_response(messages)


#[GET] /api/messages/MESSAGE_ID	Read a Single Stream Message
@restapi.route("/messages/<int:message_id>")
@login_required
def get_message(message_id):
  message = Message.query.get(message_id)
  return make_json_response(message)


#[PUT] /api/messages/MESSAGE_ID	Updating Existing Messages
@restapi.route("/messages/<int:message_id>", methods=['PUT'])
@login_required
def update_message(message_id):
  message = Message.query.get(message_id)
  d = get_params(["content"])
  message.update(d)
  db.session.commit()
  return make_json_response(message)


#[DELETE] /api/messages/MESSAGE_ID	Destroy an existing message
@restapi.route("/messages/<int:message_id>", methods=['DELETE'])
@login_required
def delete_message(message_id):
  message = Message.query.get(message_id)
  db.session.delete(message)
  db.session.commit()
  return make_response("", 204)


#
# Likes
#
# TODO: use an "objects" namespace instead to make it generic?

#[POST] /api/messages/MESSAGE_ID/likes	Liking a Message
@restapi.route("/messages/<int:message_id>/likes", methods=['POST'])
@login_required
def like_message(message_id):
  pass


#[POST] /api/comments/COMMENT_ID/likes/LIKES_ID	Liking a Comment
@restapi.route("/comments/<int:comment_id>/likes", methods=['POST'])
@login_required
def like_comment(comment_id):
  pass


#[DELETE] /api/messages/MESSAGE_ID/likes/LIKES_ID	Un-liking a Message
@restapi.route("/messages/<int:message_id>/likes/<int:like_id>", methods=['DELETE'])
@login_required
def unlike_message(message_id, like_id):
  pass


#[DELETE] /api/comments/COMMENT_ID/likes/LIKES_ID
@restapi.route("/comments/<int:comment_id>/likes/<int:like_id>", methods=['DELETE'])
@login_required
def unlike_comment(comment_id, like_id):
  pass


#
# Search
#

#[GET] /api/messages/search	Searching Messages
@restapi.route("/search/messages")
@login_required
def search_messages():
  q = request.args.get("q")
  if not q:
    return make_json_response([])
  messages = list(Message.search_query(q).all())
  return make_json_response(messages)


#[GET] /api/users/search	Search Users in Your Company
@restapi.route("/search/users")
@login_required
def search_users():
  q = request.args.get("q")
  if not q:
    return make_json_response([])
  users = list(User.search_query(q).all())
  return make_json_response(users)

#
# Activity Streams
#
@restapi.route("/feed")
@login_required
def get_feed():
  pass



"""
Default ("home") page for social apps.
"""


def register_plugin(app):
  from .views.social import social
  from .views import users, groups, sidebars  # noqa

  app.register_blueprint(social)

  # TODO: better config variable choice?
  if app.config.get("SOCIAL_REST_API"):
    from .restapi import restapi
    app.register_blueprint(restapi)



# coding=utf-8
"""
Models for the Forum.

Note: a few features are planned but not implemented yet, and are commented
out.
"""
from __future__ import absolute_import
from itertools import chain

import sqlalchemy as sa
from sqlalchemy import Column, Integer, ForeignKey, Unicode, UnicodeText
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.hybrid import hybrid_property

from abilian.core.entities import Entity, SEARCHABLE
from abilian.services.indexing.adapter import SAAdapter

from abilian.sbe.apps.communities.models import (
  Community, CommunityIdColumn, community_content
)
from abilian.sbe.apps.documents.models import BaseContent, CmisObject


class ThreadClosedError(RuntimeError):

  def __init__(self, thread):
    super(ThreadClosedError, self).__init__(
      'The thread {!r} is closed. No modification allowed on its posts: '
      'creation, edition, deletion'.format(thread),)
    self.thread = thread


@community_content
class Thread(Entity):
  """
  A thread contains conversations among forum participants.

  The discussions in a thread may be sorted in chronological order or threaded
  by reply.

  (= Thread in SIOC, Message in ICOM 1.0).
  """
  __tablename__ = 'forum_thread'

  community_id = CommunityIdColumn()
  #: The community this thread belongs to
  community = relationship(Community,
                           primaryjoin=(community_id == Community.id),
                           backref=backref('threads',
                                           cascade="all, delete-orphan"))

  #: The thread title (aka subject)
  _title = Column('title', Unicode(255), nullable=False,
                  default=u"", info=SEARCHABLE)

  # title is defined has an hybrid property to allow name <-> title sync (2 way)
  @hybrid_property
  def title(self):
    return self._title

  @title.setter
  def title(self, title):
    # set title before setting name, so that we don't enter an infinite loop
    # with _thread_sync_name_title
    self._title = title
    if self.name != title:
      self.name = title

  posts = relationship('Post',
                       primaryjoin='Thread.id == Post.thread_id',
                       order_by='Post.created_at',
                       cascade="all, delete-orphan",
                       back_populates='thread',)

  @property
  def closed(self):
    """
    True if this thread doesn't accept more posts.
    """
    return self.meta.get('abilian.sbe.forum', {}).get('closed', False)

  @closed.setter
  def closed(self, value):
    self.meta.setdefault('abilian.sbe.forum', {})['closed'] = bool(value)
    self.meta.changed()

  def create_post(self, **kw):
    if self.closed:
      raise ThreadClosedError(self)

    kw['name'] = self.name
    post = Post(**kw)
    post.thread = self
    return post


@sa.event.listens_for(Thread.name, "set", active_history=True)
def _thread_sync_name_title(entity, new_value, old_value, initiator):
  """
  Synchronize thread name -> title.
  thread.title -> name is done via hybrid_property, avoiding infinite
  loop (since "set" is received before attribute has received value)
  """
  if entity.title != new_value:
    entity.title = new_value
  return new_value


class Post(Entity):
  """
  A post is a message in a forum discussion thread.

  (= Post in DiscussionMessage in ICOM 1.0).
  """
  __tablename__ = 'forum_post'
  __indexable__ = False # content is indexed at thread level

  #: The thread this post belongs to
  thread_id = Column(ForeignKey(Thread.id), nullable=False)
  thread = relationship(Thread,
                        foreign_keys=thread_id,
                        back_populates='posts')

  #: The post this post is a reply to, if any (currently not used)
  parent_post_id = Column(ForeignKey("forum_post.id"), nullable=True)
  parent_post = relationship("Post", foreign_keys=[parent_post_id])

  #: Markup type (Markdown, Textile...)
  # TODO: markup langage selection + default
  #markup_type = Column(String, default="Markdown")

  #: Source (markup) for the post
  #body_src = Column(UnicodeText, default=u"", nullable=False)

  #: HTML rendering of the post
  body_html = Column(UnicodeText, default=u"", nullable=False)

  @hybrid_property
  def title(self):
    return self.name

  @property
  def history(self):
    return self.meta.get('abilian.sbe.forum', {}).get('history', [])


class ThreadIndexAdapter(SAAdapter):
  """
  Index a thread and its posts
  """
  @staticmethod
  def can_adapt(obj_cls):
    return obj_cls is Thread

  def get_document(self, obj):
    kw = super(ThreadIndexAdapter, self).get_document(obj)
    kw['text'] = u' '.join(chain((kw['text'],), [p.body_html for p in obj.posts]))
    return kw


# event listener to sync name with thread's name
@sa.event.listens_for(Thread.name, "set", active_history=True)
def _thread_sync_name(thread, new_value, old_value, initiator):
  """
  Synchronize name with thread's name.
  """
  if new_value == old_value:
    return new_value

  for post in thread.posts:
    post.name = new_value
  return new_value

@sa.event.listens_for(Post.thread, "set", active_history=True)
def _thread_change_sync_name(post, new_thread, old_thread, initiator):
  """
  Change name on thread change
  """
  if new_thread == old_thread or new_thread is None:
    return new_thread
  post.name = new_thread.name
  return new_thread


@sa.event.listens_for(Thread.posts, 'append')
@sa.event.listens_for(Thread.posts, 'remove')
@sa.event.listens_for(Thread.posts, 'set')
def _guard_closed_thread_collection(thread, value, *args):
  '''
  Prevent add/remove/replace posts on a closed thread
  '''
  if isinstance(thread, Post):
    thread = thread.thread
    if thread is None:
      return

  if thread.closed:
    raise ThreadClosedError(thread)

  return value



class PostAttachment(BaseContent, CmisObject):
  __tablename__ = None
  __mapper_args__ = {'polymorphic_identity': 'forum_post_attachment'}
  sbe_type = 'forum_post:attachment'

  _post_id = Column(Integer, ForeignKey(Post.id), nullable=True)
  post = relationship(Post,
                      primaryjoin=(_post_id == Post.id),
                      backref=backref('attachments', lazy='select',
                                      order_by='PostAttachment.name',
                                      cascade='all, delete-orphan'))



from ..cmis.parser import Entry

XML_ENTRY = """\
<?xml version="1.0" encoding="utf-8"?>
<entry xmlns="http://www.w3.org/2005/Atom"
  xmlns:app="http://www.w3.org/2007/app"
  xmlns:cmisra="http://docs.oasis-open.org/ns/cmis/restatom/200908/">
  <cmisra:object xmlns:cmis="http://docs.oasis-open.org/ns/cmis/core/200908/">
    <cmis:properties>
      <cmis:propertyString propertyDefinitionId="cmis:name">
        <cmis:value>Toto Titi</cmis:value>
      </cmis:propertyString>
      <cmis:propertyId propertyDefinitionId="cmis:objectTypeId">
        <cmis:value>cmis:folder</cmis:value>
      </cmis:propertyId>
    </cmis:properties>
  </cmisra:object>
  <title>Toto Titi</title>
</entry>
"""

XML_ENTRY_WITH_CONTENT = """\
<?xml version="1.0" encoding="utf-8"?>
<entry xmlns="http://www.w3.org/2005/Atom"
  xmlns:app="http://www.w3.org/2007/app"
  xmlns:cmisra="http://docs.oasis-open.org/ns/cmis/restatom/200908/">
  <cmisra:content>
    <cmisra:mediatype>text/plain</cmisra:mediatype>
    <cmisra:base64>VGVzdCBjb250ZW50IHN0cmluZw==</cmisra:base64>
  </cmisra:content>
  <cmisra:object
    xmlns:cmis="http://docs.oasis-open.org/ns/cmis/core/200908/">
    <cmis:properties>
      <cmis:propertyString propertyDefinitionId="cmis:name">
        <cmis:value>testDocument</cmis:value>
      </cmis:propertyString>
      <cmis:propertyId propertyDefinitionId="cmis:objectTypeId">
        <cmis:value>cmis:document</cmis:value>
      </cmis:propertyId>
    </cmis:properties>
  </cmisra:object>
  <title>testDocument</title>
</entry>
"""


def test_parse_folder_entry():
  e = Entry(XML_ENTRY)
  assert e.name == "Toto Titi"
  assert e.type == "cmis:folder"

def test_parse_document_entry():
  e = Entry(XML_ENTRY_WITH_CONTENT)
  assert e.name == "testDocument"
  assert e.type == "cmis:document"
  assert e.content == "Test content string"



from os.path import dirname

from lxml import etree

from ..webdav.xml import Propfind, MultiStatus
from ..webdav.constants import DAV_PROPS


def test_propfind_sample1():
  xml = open(dirname(__file__) + "/data/propfind1.xml").read()
  propfind = Propfind(xml)
  assert propfind.mode == 'prop'

def test_propfind_sample2():
  xml = open(dirname(__file__) + "/data/propfind2.xml").read()
  propfind = Propfind(xml)
  assert propfind.mode == 'prop'

def test_propfind_sample3():
  xml = open(dirname(__file__) + "/data/propfind3.xml").read()
  propfind = Propfind(xml)
  assert propfind.mode == 'allprop'

def test_empty_multistatus():
  m = MultiStatus()
  result = m.to_string()

  # Check XML is weel-formed
  etree.fromstring(result)

def test_multistatus():
  class Obj(object):
    pass
  obj = Obj()
  obj.name = "some name"
  obj.is_folder = True
  m = MultiStatus()
  m.add_response_for("http://example.com/", obj, DAV_PROPS)
  result = m.to_string()

  # Check XML is weel-formed
  etree.fromstring(result)



# coding=utf-8
"""
Lightweight integration and denormalisation using events (signals).
"""

from __future__ import absolute_import

from blinker import ANY
from abilian.core.signals import activity
from abilian.sbe.apps.documents.models import Document

from .models import Community


@activity.connect_via(ANY)
def update_community(sender, verb, actor, object, target=None):
  if isinstance(object, Community):
    object.touch()
    return

  if isinstance(target, Community):
    community = target
    community.touch()

    if isinstance(object, Document):
      if verb == 'post':
        community.document_count += 1
      elif verb == 'delete':
        community.document_count -= 1



# coding=utf-8
"""
"""
from __future__ import absolute_import

import logging
import uuid
import pkg_resources
from pathlib import Path

from abilian.core.util import fqcn


logger = logging.getLogger(__name__)
STATIC_DIR = pkg_resources.resource_filename(__name__, 'static')
LESSCSS_FILE = str(Path(STATIC_DIR, 'less', 'abilian-sbe.less'))
JS = ('js/sbe-datatable.js',
      'js/jquery.fileapi.js',
      'js/folder.js',
      'js/folder_edit.js',
      'js/folder_upload.js',
      'js/document_viewer.js',
)


class AbilianSBE(object):
  """
  Base extension required by abilian.sbe.apps
  """
  def __init__(self, app=None):
    if app is not None:
      self.init_app(app)

  def init_app(self, app):
    # False: it's ok if antivirus task was run but service couldn't get a result
    app.config.setdefault('ANTIVIRUS_CHECK_REQUIRED', False)
    app.config.setdefault('SBE_FORUM_REPLY_BY_MAIL', False)

    if FQCN in app.extensions:
      return

    app.extensions[FQCN] = self

    # register i18n
    app.extensions['babel'].add_translations('abilian.sbe')

    # sbe static assets
    app.add_static_url('abilian/sbe', STATIC_DIR, endpoint='abilian_sbe_static')
    app.extensions['webassets'].append_path(
      STATIC_DIR,
      app.static_url_path + '/abilian/sbe')

    app.register_asset('js', *JS)
    app.register_asset('css', LESSCSS_FILE)
    logger.info('Register jinja context processors')
    app.context_processor(inject_template_utils)


FQCN = fqcn(AbilianSBE)
sbe = AbilianSBE()


def inject_template_utils():
  return dict(uuid=uuid.uuid1)



"""
Some functions to retrieve activity entries.

"""
# TODO: move to the activity service ?

from flask import g, current_app
from flask_login import current_user
from werkzeug.exceptions import Forbidden
import sqlalchemy as sa
from abilian.core.extensions import db
from abilian.services.security import Admin, READ
from abilian.services.activity import ActivityEntry

from abilian.sbe.apps.communities.models import Membership
from abilian.sbe.apps.documents.models import Folder, Document


def get_recent_entries(num=20, user=None, community=None):
  AE = ActivityEntry

  # Check just in case
  if not current_user.has_role(Admin):
    if community and not community.has_member(current_user):
      raise Forbidden()

  query = AE.query.options(sa.orm.joinedload(AE.object))

  if community:
    query = query.filter(sa.or_(AE.target == g.community,
                                AE.object == g.community))
  if user:
    query = query.filter(AE.actor == user)

  # Security check
  #
  # we use communities ids instead of object because as of sqlalchemy 0.8 the
  # 'in_' operator cannot be used with relationships, only foreign keys values
  if not community and not current_user.has_role(Admin):
    M = Membership
    communities = M.query\
      .filter(M.user_id == current_user.id)\
      .values(M.community_id)

    communities = list(communities) # convert generator to list: we'll need it
                                    # twice in query filtering
    if not communities:
      return []

    query = query.filter(sa.or_(AE.target_id.in_(communities),
                                AE.object_id.in_(communities)))

  query = query.order_by(AE.happened_at.desc()).limit(1000)
  limit = min(num * 2, 100)  # get twice entries as needed, but ceil to 100
  entries = []
  deleted = False
  security = current_app.services['security']
  has_permission = security.has_permission

  for entry in query.yield_per(limit):
    if len(entries) >= num:
      break

    # Remove entries corresponding to deleted objects
    if entry.object is None:
      db.session.delete(entry)
      deleted = True
      continue

    if (isinstance(entry.object, (Folder, Document))
        and not has_permission(current_user, READ,
                               obj=entry.object,
                               inherit=True)):
        continue

    entries.append(entry)

  if deleted:
    db.session.commit()

  return entries



# coding=utf-8

from wtforms import TextAreaField, StringField
from wtforms_alchemy import model_form_factory
from flask_babel import lazy_gettext as _l
from abilian.web.forms import Form, widgets as abilian_widgets
from abilian.web.forms.fields import QuerySelect2Field
from abilian.web.forms.validators import required, optional
from abilian.web.forms.filters import strip

from abilian.sbe.apps.communities.models import Community


#from .widgets import UserPhotoInputWidget


ModelForm = model_form_factory(Form)


# FIXME
class UserProfileForm(ModelForm):
  pass


class UserProfileViewForm(UserProfileForm):
  communautes = QuerySelect2Field(
    u'Communautés d\'appartenance',
    get_label='name',
    view_widget=abilian_widgets.ListWidget(),
    query_factory=lambda: Community.query.all(),
    multiple=True,
    validators=[optional()])


class GroupForm(Form):
  name = StringField(
    _l("Name"),
    filters=(strip,),
    validators=[required(message=_l("Name is required."))])

  description = TextAreaField(_l("Description"))



# coding=utf-8

from abilian.sbe.testing import BaseTestCase
from sqlalchemy.exc import IntegrityError

from abilian.core.extensions import db

from ..models import Folder, Document


class TestService(BaseTestCase):

  def setUp(self):
    super(TestService, self).setUp()
    self.repository = self.app.extensions['content_repository']
    self.root = Folder(title=u"")
    self.session.add(self.root)
    self.session.flush()

  def test_create_doc(self):
    root = self.root
    doc = root.create_document(u"doc")
    db.session.flush()
    assert doc.path == "/doc"
    assert len(root.children) == 1
    assert not doc.is_root_folder

  def test_create_folder(self):
    root = self.root
    folder = root.create_subfolder(u"folder")
    db.session.flush()
    self.assertEquals(folder.title, u'folder')
    self.assertEquals(folder.name, u'folder')
    assert folder.path == "/folder"
    assert len(root.children) == 1
    assert len(folder.children) == 0
    assert not folder.is_root_folder

  def test_move(self):
    root = self.root
    doc = root.create_document(u"doc")
    folder = root.create_subfolder(u"folder")

    self.repository.move_object(doc, folder, u"newdoc")
    assert doc in folder.children
    assert doc not in root.children
    assert len(root.children) == 1
    assert len(folder.children) == 1
    assert doc.title == "newdoc"
    assert doc.path == "/folder/newdoc"

  def test_copy(self):
    root = self.root
    doc = root.create_document(u"doc")
    folder = root.create_subfolder(u"folder")

    doc_copy = self.repository.copy_object(doc, folder, u"copydoc")

    assert len(root.children) == 2
    assert len(folder.children) == 1
    self.assertEquals(doc_copy.title, "copydoc")
    self.assertEquals(doc_copy.name, "copydoc")
    assert doc_copy.path == "/folder/copydoc"
    assert doc_copy in folder.children

  def test_copy_nested_folders(self):
    root = self.root
    folder1 = root.create_subfolder(u"folder1")
    folder2 = root.create_subfolder(u"folder2")
    subfolder = folder1.create_subfolder(u"subfolder")

    folder1_copy = self.repository.copy_object(folder1, folder2)

    assert len(root.children) == 2
    assert folder1 in root.children
    assert folder2 in root.children

    assert len(folder1.children) == 1
    assert subfolder in folder1.children

    assert len(folder2.children) == 1
    assert folder1_copy in folder2.children

  def test_move_nested_folders(self):
    root = self.root
    folder1 = root.create_subfolder(u"folder1")
    folder2 = root.create_subfolder(u"folder2")
    subfolder = folder1.create_subfolder(u"subfolder")  # noqa

    self.repository.move_object(folder1, folder2)

    assert len(root.children) == 1
    assert len(folder1.children) == 1

  def test_rename(self):
    root = self.root
    folder = root.create_subfolder(u"folder")
    doc = folder.create_document(u"doc")

    self.repository.rename_object(doc, u"doc1")
    self.assertEquals(doc.title, 'doc1')
    self.assertEquals(doc.name, 'doc1')
    assert doc.path == '/folder/doc1'

    self.repository.rename_object(folder, u"folder1")
    self.assertEquals(folder.title, 'folder1')
    self.assertEquals(folder.name, 'folder1')
    assert folder.path == '/folder1'
    assert doc.path == '/folder1/doc1'

  def test_delete(self):
    root = self.root

    doc = root.create_document(u"doc")
    folder = root.create_subfolder(u"folder")
    self.session.flush()

    assert doc.parent == root
    assert folder.parent == root
    assert doc in root.children
    assert folder in root.children
    assert not doc.is_root_folder
    assert not folder.is_root_folder

    doc_id = doc.id
    folder_id = folder.id

    self.repository.delete_object(doc)
    self.repository.delete_object(folder)
    assert len(root.children) == 0
    self.session.flush()

    assert Folder.query.get(folder_id) is None
    assert Document.query.get(doc_id) is None

    # test delete tree
    folder = root.create_subfolder(u"folder")
    sub = folder.create_subfolder(u'subfolder')  # noqa
    doc = folder.create_document(u"doc")  # noqa
    self.session.flush()

    self.repository.delete_object(folder)
    assert len(root.children) == 0
    assert Folder.query.all() == [root]
    assert Document.query.all() == []

  def test_no_duplicate_name(self):
    root = self.root
    root.create_subfolder(u'folder_1')
    root.create_subfolder(u'folder_1')
    try:
      self.session.flush()
      self.fail('Could create folders with duplicate name')
    except IntegrityError:
        pass



"""
Provides content for sidebars, accessible via the 'g' object.

TODO: make it smarter (direct access from 'g' using lazy objects?) and
cacheable.
"""

from datetime import datetime, timedelta
from flask import g

from abilian.core.models.subjects import User
from flask.ext.login import current_user
from abilian.sbe.apps.communities.models import Community, Membership
from abilian.sbe.apps.social.views.social import social


class Sidebars(object):

  @property
  def latest_visitors(self):
    return User.query.filter(User.last_active != None) \
      .order_by(User.last_active.desc()) \
      .limit(15) \
      .all()

  @property
  def active_visitor_count(self):
    one_minute_ago = (datetime.utcnow() - timedelta(0, 60))
    return User.query.filter(
      User.last_active > one_minute_ago).count()

  @property
  def my_communities(self):
    query = Community.query
    query = query.order_by(Community.last_active_at.desc())
    if not current_user.has_role('admin'):
      # Filter with permissions
      query = query.join(Membership).filter(Membership.user == current_user)
    return query.limit(10).all()

  @property
  def all_communities(self):
    # TODO: limit
    return []
    #return Community.query.all()


@social.before_request
def inject_sidebars():
  g.sidebars = Sidebars()



# coding=utf-8
"""
Indexing related utilities for Folder, Documents
"""
from __future__ import absolute_import

import sqlalchemy as sa
from flask import current_app

from abilian.core.entities import Entity

from .models import CmisObject

def reindex_tree(obj):
  """
  Schedule reindexing `obj` and all of its descendants.

  Generally needed to update indexed security.
  """
  assert isinstance(obj, CmisObject)
  
  svc = current_app.services['indexing']
  if not svc.running:
    return
    
  descendants = sa.select([CmisObject.id, CmisObject._parent_id])\
                  .where(CmisObject._parent_id == sa.bindparam('ancestor_id'))\
                  .cte(name='descendants', recursive=True)
  da = descendants.alias()
  CA = sa.orm.aliased(CmisObject)
  d_ids = sa.select([CA.id, CA._parent_id])
  descendants = descendants.union_all(d_ids.where(CA._parent_id == da.c.id))
  session = sa.orm.object_session(obj) or current_app.db.session()
  
  # including ancestor_id in entity_ids_q will garantee at least 1 value for the
  # "IN" predicate; otherwise when using sqlite (as during tests...)
  # 'ResourceClosedError' will be raised.
  #
  # as an added bonus, "obj" will also be in query results, thus will be added
  # in "to_update" without needing to do it apart.
  entity_ids_q = sa.union(sa.select([descendants.c.id]),
                          sa.select([sa.bindparam('ancestor_id')]))
  q = session.query(Entity)\
             .filter(Entity.id.in_(entity_ids_q)) \
             .options(sa.orm.noload('*')) \
             .params(ancestor_id=obj.id)

  to_update = svc.app_state.to_update
  key = 'changed'
  
  for item in q.yield_per(1000):
    to_update.append((key, item))
  



# coding=utf-8
"""
Forum module
"""
from __future__ import absolute_import

from abilian.sbe.extension import sbe


def register_plugin(app):
  app.config.setdefault('SBE_FORUM_REPLY_BY_MAIL', False)
  app.config.setdefault('MAIL_ADDRESS_TAG_CHAR', u'+')
  app.config.setdefault('INCOMING_MAIL_USE_MAILDIR', False)
  sbe.init_app(app)
  from .views import forum
  from .actions import register_actions
  from .models import ThreadIndexAdapter
  from .commands import manager
  from . import tasks

  forum.record_once(register_actions)
  app.register_blueprint(forum)
  app.services['indexing'].adapters_cls.insert(0, ThreadIndexAdapter)
  tasks.init_app(app)

  if app.script_manager:
    app.script_manager.add_command('forum', manager)



"""
The blueprint for this app.
"""

from flask import Blueprint, redirect, url_for, render_template
from abilian.core import signals
from abilian.core.util import get_params
from abilian.core.extensions import db

from abilian.sbe.apps.wall.presenters import ActivityEntryPresenter
from abilian.sbe.apps.wall.util import get_recent_entries
from ..models import Message


__all__ = ['social']

MAX_LAST_USERS = 15
MAX_MESSAGES = 15

social = Blueprint("social", __name__,
                   url_prefix='/social', template_folder='../templates')
route = social.route


@route("/")
def home():
  ctx = {}
  entries = get_recent_entries(num=50)
  entries = ActivityEntryPresenter.wrap_collection(entries)
  ctx['activity_entries'] = entries

  return render_template("social/home.html", **ctx)


@route("/stream/<stream_name>")
def stream(stream_name):
  pass


@route("/", methods=['POST'])
def share():
  # TODO: better error control / feedback
  d = get_params(Message.__editable__)
  if not d.get('content'):
    return redirect(url_for(".home"))

  message = Message(**d)
  db.session.add(message)

  #tags = message.tags
  #values = [ {'tag': tag, 'message_id': message.id} for tag in tags ]
  #db.engine.execute(tagging.insert(), values)

  db.session.commit()
  return redirect(url_for(".home"))



# coding=utf-8
"""
Entity objects for the Document Management applications.

TODO: move to an independent service / app.
"""
from __future__ import absolute_import

import logging
import threading
import hashlib
import itertools
import os
import mimetypes
import pkg_resources
import uuid

from flask import json, g, url_for, current_app
from flask_login import current_user

import sqlalchemy as sa
from sqlalchemy import event
from sqlalchemy.orm.session import Session
from sqlalchemy.orm import relationship, backref, foreign, remote
from sqlalchemy.schema import Column, ForeignKey, UniqueConstraint
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.types import UnicodeText, Text, Integer

import whoosh.fields as wf
from whoosh.analysis import CharsetFilter, RegexTokenizer, LowercaseFilter
from whoosh.support.charset import accent_map

from abilian.core.models import NOT_AUDITABLE, SEARCHABLE
from abilian.core.models.subjects import User, Group
from abilian.core.models.blob import Blob
from abilian.core.entities import db, Entity
from abilian.services.conversion import converter
from abilian.services.security import InheritSecurity, security, Anonymous, \
    Admin
from abilian.services.indexing import indexable_role

from .lock import Lock
from . import tasks


logger = logging.getLogger(__package__)

__all__ = ['db', 'Folder', 'Document', 'BaseContent', 'icon_for']

#: A Whoosh analyzer that folds accents and case.
accent_folder = (RegexTokenizer() | LowercaseFilter()
                 | CharsetFilter(accent_map))

ICONS_FOLDER = pkg_resources.resource_filename('abilian.sbe',
                                               'static/fileicons')


def icon_url(filename):
  return url_for('abilian_sbe_static', filename='fileicons/' + filename)


def icon_exists(filename):
  fullpath = os.path.join(ICONS_FOLDER, filename)
  return os.path.isfile(fullpath)


#
# Domain classes
#
class CmisObject(Entity, InheritSecurity):
  """
  (Abstract) Base class for CMIS objects.
  """
  # normally set by communities.models.community_content,
  # but we have to fix a circ.dep to use it
  is_community_content = True

  __tablename__ = 'cmisobject'
  __indexable__ = False
  __indexation_args__ = {}
  __indexation_args__.update(Entity.__indexation_args__)
  index_to = Entity.__indexation_args__.setdefault('index_to', ())
  index_to += (('community.id', ('community_id',)),
               ('community.slug', ('community_slug',)),
               )
  __indexation_args__['index_to'] = index_to
  del index_to

  _title = Column('title', UnicodeText, nullable=False, default=u"")
  description = Column(UnicodeText, nullable=False, default=u"",
                       info=SEARCHABLE | dict(index_to=('description', 'text')))

  _parent_id = Column(Integer, ForeignKey('cmisobject.id'), nullable=True)

  # no duplicate name in same folder
  __table_args__ = (UniqueConstraint('_parent_id', 'title'),)

  # Set in concrete classes
  sbe_type = None

  # Convenience default values
  content_length = 0

  def __init__(self, *args, **kwargs):
    # ensure 'title' prevails over 'name'
    if 'title' in kwargs and 'name' in kwargs:
      title = kwargs.get('title')
      name = kwargs.get('name')

      if title is None:
        del kwargs['title']
      elif name != title:
        kwargs['name'] = title

    Entity.__init__(self, *args, **kwargs)

  # title is defined has an hybrid property to allow 2 way sync name <-> title
  @hybrid_property
  def title(self):
    return self._title

  @title.setter
  def title(self, title):
    # set title before setting name, so that we don't enter an infinite loop
    # with _cmis_sync_name_title
    self._title = title
    if self.name != title:
      self.name = title

  def clone(self, title=None, parent=None):
    if not title:
      title = self.title
    new_obj = self.__class__(title=title, name=title, parent=parent)
    state = vars(self)
    for k, v in state.items():
      if not k.startswith("_") and k not in ['uid', 'id', 'parent', 'title',
                                             'name', 'path', 'subfolders',
                                             'documents']:
        setattr(new_obj, k, v)
    if self.parent:
      assert self in self.parent.children
    return new_obj

  @property
  def path(self):
    if self.parent:
      return self.parent.path + u"/" + self.title
    else:
      return u""

  @property
  def is_folder(self):
    return self.sbe_type == 'cmis:folder'

  @property
  def is_document(self):
    return self.sbe_type == 'cmis:document'

  @property
  def is_root_folder(self):
    return self._parent_id is None

  @property
  def community(self):
    if not self.is_folder:
      return self.parent and self.parent.community

    if self._community:
      return self._community

    return self.parent and self.parent.community


@event.listens_for(CmisObject.name, "set", propagate=True, active_history=True)
def _cmis_sync_name_title(entity, new_value, old_value, initiator):
  """
  Synchronize CmisObject name -> title.
  CmisObject.title -> name is done via hybrid_property, avoiding infinite
  loop (since "set" is received before attribute has received value)
  """
  if entity.title != new_value:
    entity.title = new_value
  return new_value


class PathAndSecurityIndexable(object):
  """
  Mixin for folder and documents indexation
  """
  __indexation_args__ = dict(
    index_to=(
      ('_indexable_parent_ids', ('parent_ids',)),
      ('_indexable_roles_and_users', ('allowed_roles_and_users',)),
    ),
  )

  def _iter_to_root(self, skip_self=False):
    obj = self if not skip_self else self.parent
    while obj:
      yield obj
      obj = obj.parent

  @property
  def _indexable_parent_ids(self):
    """
    returns a string made of ids separated by a slash: "/1/3/4/5", "5" being
    self.parent.id.
    """
    ids = [unicode(obj.id) for obj in self._iter_to_root(skip_self=True)]
    return u'/' + u'/'.join(reversed(ids))

  @property
  def _indexable_roles_and_users(self):
    """
    returns a string made of type:id elements, like "user:2 group:1 user:6"
    """
    iter_from_root = reversed(list(self._iter_to_root()))
    if self.parent:
      # skip root folder only on non-root folder!
      iter_from_root.next()
    allowed = set(o[0] for o in security.get_role_assignements(iter_from_root.next()))

    for obj in iter_from_root:
      if obj.inherit_security:
        continue
      obj_allowed = set(o[0] for o in security.get_role_assignements(obj))

      if Anonymous in obj_allowed:
        continue

      parent_allowed = allowed
      # pure intersection: users and groups in both are preserved
      allowed = allowed & obj_allowed
      remaining = parent_allowed - obj_allowed
      # find users who can access 'obj' because of their group memberships
      # 1. extends groups in obj_allowed with their actual member list
      extended_allowed = set(
        itertools.chain(*(p.members if isinstance(p, Group) else (p,)
                        for p in obj_allowed)))

      # 2. remaining_users are users explicitly listed in parents but not on
      # obj. Are they in a group?
      remaining_users = set(o for o in remaining if isinstance(o, User))
      allowed |= (remaining_users & extended_allowed)

      # remaining groups: find if some users are eligible
      remaining_groups_members = set(
        itertools.chain(*(p.members for p in remaining
                          if isinstance(p, Group)))
      )
      allowed |= remaining_groups_members - extended_allowed

    # admin role is always granted access
    allowed.add(Admin)
    return u' '.join(indexable_role(p) for p in allowed)


class Folder(CmisObject, PathAndSecurityIndexable):
  __tablename__ = None
  sbe_type = 'cmis:folder'

  __indexable__ = True
  __indexation_args__ = {}
  __indexation_args__.update(CmisObject.__indexation_args__)
  index_to = tuple()
  index_to += CmisObject.__indexation_args__.setdefault('index_to', ())
  index_to += PathAndSecurityIndexable.__indexation_args__.setdefault('index_to', ())
  __indexation_args__['index_to'] = index_to
  del index_to

  _indexable_roles_and_users = PathAndSecurityIndexable.\
                               _indexable_roles_and_users

  parent = relationship(
    "Folder",
    primaryjoin=(lambda: foreign(CmisObject._parent_id) == remote(Folder.id)),
    backref=backref('subfolders',
                    lazy="joined",
                    order_by='Folder.title',
                    cascade='all, delete-orphan'))

  @property
  def icon(self):
   return icon_url('folder.png')

  @property
  def children(self):
    return self.subfolders + self.documents

  @property
  def document_count(self):
    count = len(self.documents)
    for f in self.subfolders:
      count += f.document_count
    return count

  @property
  def depth(self):
    if self.parent is None:
      return 0
    else:
      return self.parent.depth + 1

  def create_subfolder(self, title):
    subfolder = Folder(title=title, parent=self)
    assert subfolder in self.children
    return subfolder

  def create_document(self, title):
    doc = Document(title=title, parent=self)
    assert doc.parent == self
    assert doc in self.children
    return doc

  def get_object_by_path(self, path):
    assert path.startswith("/")
    assert "//" not in path

    if path == "/":
      return self

    path_segments = path[1:].split("/")
    obj = self
    try:
      for name in path_segments[:]:
        obj = filter(lambda x: x.title == name, obj.children)[0]
      return obj
    except IndexError:
      return None

  def __repr__(self):
    return "<%s.%s id=%r name=%r path=%r at 0x%x>" % (
      self.__class__.__module__, self.__class__.__name__,
      self.id, self.title, self.path, id(self)
    )

  #
  # Security related methods
  #
  @property
  def filtered_children(self):
    return security.filter_with_permission(g.user, "read", self.children,
                                           inherit=True)

  @property
  def filtered_subfolders(self):
    return security.filter_with_permission(g.user, "read", self.subfolders,
                                           inherit=True)

  def get_local_roles_assignments(self):
    local_roles_assignments = security.get_role_assignements(self)
    # local_roles_assignments = sorted(local_roles_assignments,
    #                                  key=lambda u: (u[0].last_name.lower(),
    #                                                 u[0].first_name.lower()))
    return local_roles_assignments

  def get_inherited_roles_assignments(self):
    if self.parent.is_root_folder:
      return []

    roles = self.parent.get_local_roles_assignments()
    inherited_roles = (self.parent.get_inherited_roles_assignments()
                       if self.parent.inherit_security
                       else [])

    assignments = set(itertools.chain(roles, inherited_roles))

    def key(x):
      principal = x[0]
      if isinstance(principal, User):
        # Defensive programming here, this shouldn't happen actually
        last_name = principal.last_name or ""
        first_name = principal.first_name or ""
        return last_name.lower(), first_name.lower()
      elif isinstance(principal, Group):
        return principal.name
      else:
        raise Exception("Bad class here: %s" % type(principal))

    return sorted(assignments, key=key)

  def members(self):
    local_roles = self.get_local_roles_assignments()
    inherited_roles = (self.get_inherited_roles_assignments()
                       if self.inherit_security else [])

    def _iter_users(roles):
      for principal, user in roles:
        if isinstance(principal, User):
          yield principal
        else:
          for user in itertools.chain(principal.members, principal.admins):
            yield user

    members = set(_iter_users(itertools.chain(local_roles, inherited_roles)))
    members = sorted(members, key=lambda u: (u.last_name, u.first_name))
    return members


class BaseContent(CmisObject):
  """ A base class for cmisobject with an attached file
  """
  __tablename__ = None

  _content_id = Column(Integer, db.ForeignKey(Blob.id))
  content_blob = relationship(Blob, cascade='all, delete',
                              foreign_keys=[_content_id])

  #: md5 digest (BTW: not sure they should be part of the public API).
  content_digest = Column(Text)

  #: size (in bytes) of the content blob.
  content_length = Column(
    Integer, default=0, nullable=False, server_default=sa.text('0'),
    info=dict(searchable=True,
              index_to=(('content_length', wf.NUMERIC(stored=True)),),)
  )

  #: MIME type of the content stream.
  # TODO: normalize mime type?
  content_type = Column(
    Text, default="application/octet-stream",
    info=dict(searchable=True,
              index_to=(('content_type', wf.ID(stored=True)),)),
  )

  @property
  def content(self):
    return self.content_blob.value

  @content.setter
  def content(self, value):
    assert isinstance(value, str)
    self.content_blob = Blob()
    self.content_blob.value = value
    self.content_length = len(value)

  def set_content(self, content, content_type=None):
    new_digest = hashlib.md5(content).hexdigest()
    if new_digest == self.content_digest:
      return

    self.content_digest = new_digest
    self.content = content
    content_type = self.find_content_type(content_type)
    if content_type:
      self.content_type = content_type

  def find_content_type(self, content_type=None):
    """ Find possibly more appropriate content_type for this instance.

    if `content_type` is a binary one, try to find a better one based on
    content name so that 'xxx.pdf' is not flagged as binary/octet-stream for
    example
    """
    if not content_type or content_type in ('application/octet-stream',
                                            'binary/octet-stream',
                                            'application/binary'):
      # absent or generic content type: try to find something more useful to be
      # able to do preview/indexing/...
      guessed_content_type = mimetypes.guess_type(self.title, strict=False)[0]
      if (guessed_content_type
          and guessed_content_type != 'application/vnd.ms-office.activeX'):
        # mimetypes got an update: "random.bin" would be guessed as
        # 'application/vnd.ms-office.activeX'... not so useful in a document
        # repository
        content_type = guessed_content_type

    return content_type

  @property
  def icon(self):
    icon = icon_for(self.content_type)

    if not icon.endswith("/bin.png"):
      return icon

    # Try harder, just in case. XXX: Could be probably removed later when we are
    # sure that all our bases are covered.
    if "." not in self.title:
      return icon_url('bin.png')

    suffix = self.title.split(".")[-1]
    icon = u'{}.png'.format(suffix)
    if icon_exists(icon):
      return icon_url(icon)
    else:
      return icon_url('bin.png')


class Document(BaseContent, PathAndSecurityIndexable):
  """A document, in the CMIS sense.
  """
  __tablename__ = None

  __indexable__ = True
  __indexation_args__ = {}
  __indexation_args__.update(BaseContent.__indexation_args__)
  index_to = tuple()
  index_to += BaseContent.__indexation_args__.setdefault('index_to', ())
  index_to += PathAndSecurityIndexable.__indexation_args__.setdefault('index_to', ())
  index_to += (('text', ('text',)),)
  __indexation_args__['index_to'] = index_to
  del index_to

  _indexable_roles_and_users = PathAndSecurityIndexable.\
                               _indexable_roles_and_users

  parent = relationship(
    "Folder",
    primaryjoin=(foreign(CmisObject._parent_id) == remote(Folder.id)),
    backref=backref('documents',
                    lazy="joined",
                    order_by='Document.title',
                    cascade='all, delete-orphan'),
  )

  PREVIEW_SIZE = 700

  @property
  def preview_size(self):
    return self.PREVIEW_SIZE

  def has_preview(self, size=None, index=0):
    if size is None:
      size = self.PREVIEW_SIZE

    return converter.has_image(self.content_digest, self.content_type, index,
                               size)

  @property
  def digest(self):
    """Alias for content_digest."""
    return self.content_digest

  _text_id = Column(Integer, db.ForeignKey(Blob.id), info=NOT_AUDITABLE)
  text_blob = relationship(Blob, cascade='all, delete', foreign_keys=[_text_id])

  _pdf_id = Column(Integer, db.ForeignKey(Blob.id), info=NOT_AUDITABLE)
  pdf_blob = relationship(Blob, cascade='all, delete', foreign_keys=[_pdf_id])

  _preview_id = Column(Integer, db.ForeignKey(Blob.id), info=NOT_AUDITABLE)
  preview_blob = relationship(Blob, cascade='all, delete',
                              foreign_keys=[_preview_id])

  language = Column(Text,
                    info=dict(searchable=True,
                              index_to=[('language', wf.ID(stored=True))]),)
  size = Column(Integer)
  page_num = Column(Integer, default=1)

  #FIXME: use Entity.meta instead
  #: Stores extra metadata as a JSON column
  extra_metadata_json = Column(UnicodeText, info=dict(auditable=False))

  sbe_type = 'cmis:document'

  # antivirus status
  def ensure_antivirus_scheduled(self):
    if not self.antivirus_required:
      return True

    if current_app.config.get('CELERY_ALWAYS_EAGER', False):
      async_conversion(self)
      return True

    task_id = self.content_blob.meta.get('antivirus_task_id')
    if task_id is not None:
      res = tasks.process_document.AsyncResult(task_id)
      if not res.failed():
        # success, or pending or running
        return True

    # schedule a new task
    self.content_blob.meta['antivirus_task_id'] = str(uuid.uuid4())
    async_conversion(self)

  @property
  def antivirus_scanned(self):
    """
    True if antivirus task was run, even if antivirus didn't return a result
    """
    return self.content_blob and 'antivirus' in self.content_blob.meta

  @property
  def antivirus_status(self):
    """
    True: antivirus has scanned file: no virus
    False: antivirus has scanned file: virus detected
    None: antivirus task was run, but antivirus didn't return a result
    """
    return self.content_blob and self.content_blob.meta.get('antivirus')

  @property
  def antivirus_required(self):
    """
    True if antivirus doesn't need to be run
    """
    required = current_app.config['ANTIVIRUS_CHECK_REQUIRED']
    return required and (not self.antivirus_scanned or
                         self.antivirus_status is None)

  @property
  def antivirus_ok(self):
    """
    True if user can safely access document content
    """
    required = current_app.config['ANTIVIRUS_CHECK_REQUIRED']
    if required:
      return self.antivirus_status is True

    return self.antivirus_status is not False

  # R/W properties
  @BaseContent.content.setter
  def content(self, value):
    BaseContent.content.fset(self, value)
    self.content_blob.meta['antivirus_task_id'] = str(uuid.uuid4())
    self.pdf_blob = None
    self.text_blob = None

  def set_content(self, content, content_type=None):
    super(Document, self).set_content(content, content_type)
    async_conversion(self)

  @property
  def pdf(self):
    return self.pdb_blob and self.pdf_blob.value

  @pdf.setter
  def pdf(self, value):
    assert isinstance(value, str)
    self.pdf_blob = Blob()
    self.pdf_blob.value = value

  # `text` is an Unicode value.
  @property
  def text(self):
    return (self.text_blob.value.decode("utf8")
            if self.text_blob is not None
            else u'')

  @text.setter
  def text(self, value):
    assert isinstance(value, unicode)
    self.text_blob = Blob()
    self.text_blob.value = value.encode("utf8")

  @property
  def extra_metadata(self):
    if not hasattr(self, '_extra_metadata'):
      if self._extra_metadata is not None:
        self._extra_metadata = json.loads(self.extra_metadata_json)
      else:
        self._extra_metadata = None
    return self._extra_metadata

  @extra_metadata.setter
  def extra_metadata(self, extra_metadata):
    self._extra_metadata = extra_metadata
    self.extra_metadata_json = unicode(json.dumps(extra_metadata))

  # TODO: or use SQLAlchemy alias?
  @property
  def file_name(self):
    return self.title

  def __repr__(self):
    return "<Document id=%r name=%r path=%r content_length=%d at 0x%x>" % (
      self.id, self.title, self.path, self.content_length, id(self),
    )

  # locking management; used for checkin/checkout - this could be generalized to
  # any entity
  @property
  def lock(self):
    """
    :returns: either `None` if no lock or current lock is expired; either the
    current valid :class:`Lock` instance.
    """
    lock = self.meta.setdefault('abilian.sbe.documents', {}).get('lock')
    if lock:
      lock = Lock(**lock)
      if lock.expired:
        lock = None

    return lock

  @lock.setter
  def lock(self, user):
    """
    Allow to do `document.lock = user` to set a lock for user

    If user is None, the lock is released
    """
    if user is None:
      del self.lock
      return

    self.set_lock(user=user)

  @lock.deleter
  def lock(self):
    """
    Remove lock, if any. `del document.lock` can be safely done even if no lock
    is set.
    """
    meta = self.meta.setdefault('abilian.sbe.documents', {})
    if 'lock' in meta:
      del meta['lock']
      self.meta.changed()

  def set_lock(self, user=None):
    if user is None:
      user = current_user

    lock = self.lock
    if lock and not lock.is_owner(user=user):
      raise RuntimeError('This document is already locked by another user')

    meta = self.meta.setdefault('abilian.sbe.documents', {})
    lock = Lock.new()
    meta['lock'] = lock.as_dict()
    self.meta.changed()




def icon_for(content_type):
  for extension, mime_type in mimetypes.types_map.items():
    if mime_type == content_type:
      extension = extension[1:]
      icon = "%s.png" % extension
      if icon_exists(icon):
        return icon_url(icon)

  return icon_url('bin.png')

# Async conversion
_async_data = threading.local()


def _get_documents_queue():
  if not hasattr(_async_data, 'documents'):
    _async_data.documents = []
  return _async_data.documents


def async_conversion(document):
  _get_documents_queue().append(
    (document, document.content_blob.meta.get('antivirus_task_id')),
  )


def _trigger_conversion_tasks(session):
  if (session is not db.session()  # this commit is not from the application session
      or session.transaction.nested):  # inside a sub-transaction: not yet written in DB
    return

  document_queue = _get_documents_queue()
  while document_queue:
    doc, task_id = document_queue.pop()
    if doc.id:
      tasks.process_document.apply_async((doc.id,), task_id=task_id)


def setup_listener():
  mark_attr = '__abilian_sa_listening'
  if getattr(_trigger_conversion_tasks, mark_attr, False):
    return

  event.listen(Session, "after_commit", _trigger_conversion_tasks)
  setattr(_trigger_conversion_tasks, mark_attr, True)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import bleach
from wtforms import TextAreaField, StringField, BooleanField

from abilian.i18n import _l
from abilian.web.forms import Form, RichTextWidget
from abilian.web.forms.validators import required, optional
from abilian.web.forms.filters import strip
from abilian.web.forms.fields import FileField


ALLOWED_TAGS = [
    'a',
    'abbr',
    'acronym',
    'b',
    'blockquote',
    'br',
    'code',
    'em',
    'i',
    'li',
    'ol',
    'strong',
    'ul',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'p',
    'u',
    'img',
]

ALLOWED_ATTRIBUTES = {
  '*': ['title'],
  'p': ['style'],
  'a': ['href', 'title'],
  'abbr': ['title'],
  'acronym': ['title'],
  'img': ['src', 'alt', 'title'],
}

ALLOWED_STYLES = [
  'text-align',
]

WIDGET_ALLOWED = {}
for attr in ALLOWED_TAGS:
  allowed = ALLOWED_ATTRIBUTES.get(attr, True)
  if not isinstance(allowed, bool):
    allowed = {tag: True for tag in allowed}
  WIDGET_ALLOWED[attr] = allowed


# instantiate this one before PostForm fields, so that it is listed first
# when Threadform is displayed
_TITLE_FIELD = StringField(label=_l(u"Title"),
                          filters=(strip,),
                          validators=[required()])

class BasePostForm(Form):
  message = TextAreaField(label=_l("Message"),
                          widget=RichTextWidget(allowed_tags=WIDGET_ALLOWED),
                          filters=(strip,),
                          validators=[required()])
  attachments = FileField(label=_l(u'Attachments'), multiple=True,
                          validators=[optional()])

  def validate_message(self, field):
    field.data = bleach.clean(field.data, tags=ALLOWED_TAGS,
                              attributes=ALLOWED_ATTRIBUTES,
                              styles=ALLOWED_STYLES,
                              strip=True)


class PostForm(BasePostForm):
  send_by_email = BooleanField(label=_l(u"Send by email?"), default=True)


class ThreadForm(PostForm):
  title = _TITLE_FIELD


class PostEditForm(BasePostForm):
  reason = StringField(label=_l(u'Reason'),
                       description=_l(u'Description of your edit'),
                       filters=(strip,),
                       validators=(optional(),),)



"""Register extensions as a plugin.

NOTE: panels are currently loaded and registered manually. This may change
in the future.
"""
from abilian.sbe.extension import sbe

def register_plugin(app):
  sbe.init_app(app)
  from .main import blueprint
  app.register_blueprint(blueprint)



# coding=utf-8
from flask import url_for
from flask_babel import lazy_gettext as _l
from abilian.web.action import actions, Action, ModalActionMixin, FAIcon

from abilian.sbe.apps.communities.actions import CommunityEndpoint


class WikiPageAction(Action):
  Endpoint = CommunityEndpoint

  def pre_condition(self, context):
    page = context.get('object')
    return bool(page)

  def url(self, context=None):
    if self._url:
      return self._url
    else:
      page = context.get('object')
      kw = self.endpoint.get_kwargs()
      kw['title'] = page.title
      return url_for(self.endpoint.name, **kw)


class WikiPageModalAction(ModalActionMixin, WikiPageAction):
  pass


class WikiAction(Action):
  Endpoint = CommunityEndpoint


_actions = (
  WikiPageAction('wiki:page', 'view', _l(u'View'),
                 endpoint='.page', icon='eye-open'),

  WikiPageAction('wiki:page', 'edit', _l(u'Edit'),
                 endpoint='.page_edit', icon='pencil'),
  WikiPageModalAction('wiki:page', 'upload_attachment',
                      _l(u'Upload an attachment'),
                      url='#upload-files', icon='plus'),
  WikiPageAction('wiki:page', 'source', _l(u'Source'),
                 endpoint='.page_source', icon=FAIcon('code')),
  WikiPageAction('wiki:page', 'changes', _l(u'Changes'),
                 endpoint='.page_changes', icon='time'),
  WikiPageModalAction('wiki:page', 'delete', _l(u'Delete'),
                      url='#modal-delete', icon='trash'),
  WikiAction('wiki:global', 'new', _l(u'New page'),
             endpoint='.page_new', icon='plus'),
  WikiAction('wiki:global', 'pages', _l(u'All pages'),
             endpoint='.wiki_pages', icon='list'),
  WikiAction('wiki:global', 'help', _l(u'Syntax help'),
             endpoint='.wiki_help', icon='info-sign'),
)


def register_actions(state):
  if not actions.installed(state.app):
    return

  with state.app.app_context():
    actions.register(*_actions)



from __future__ import absolute_import
from unittest import skip

from flask import url_for

from abilian.sbe.testing import BaseTestCase


class TestsViews(BaseTestCase):
  no_login = True

  @skip("Doesn't work. Needs a real user.")
  def test_crm_notifications(self):
    response = self.client.get(url_for("preferences.crm_notifications"))
    self.assert_200(response)



# coding=utf-8
"""
This panel manages user setting for email reminders related to the SBE
(social netowking) app.
"""
from flask import render_template, redirect, url_for, request, flash
from flask import current_app as app
from werkzeug.exceptions import InternalServerError
from wtforms import BooleanField

from abilian.core.extensions import db
from abilian.i18n import _, _l
from abilian.services.preferences.panel import PreferencePanel
from abilian.web import csrf
from abilian.web.forms import Form, widgets


class SbeNotificationsForm(Form):
  daily = BooleanField(
    label=_(u"Receive by email a daily digest of activities in your communities"),
    widget=widgets.BooleanWidget(on_off_mode=True))


class SbeNotificationsPanel(PreferencePanel):
  id = 'sbe_notifications'
  label = _l(u'Community notifications')

  def is_accessible(self):
    return True

  def get(self):
    # Manual security check, should be done by the framework instead.
    if not self.is_accessible():
      raise InternalServerError()

    preferences = app.services['preferences']
    data = {}
    prefs = preferences.get_preferences()

    for k, v in prefs.items():
      if k.startswith("sbe:notifications:"):
        data[k[18:]] = v

    form = SbeNotificationsForm(formdata=None, prefix=self.id, **data)
    return render_template("preferences/sbe_notifications.html",
                           form=form)

  @csrf.protect
  def post(self):
    # Manual security check, should be done by the framework instead.
    if not self.is_accessible():
      raise InternalServerError()

    if request.form['_action'] == 'cancel':
      return redirect(url_for(".sbe_notifications"))
    form = SbeNotificationsForm(request.form, prefix=self.id)
    if form.validate():
      preferences = app.services['preferences']
      for field_name, field in form._fields.items():
        if field is form.csrf_token:
          continue
        key = 'sbe:notifications:{}'.format(field_name)
        value = field.data
        preferences.set_preferences(**{key: value})

      db.session.commit()
      flash(_(u"Preferences saved."), "info")
      return redirect(url_for(".sbe_notifications"))
    else:
      return render_template("preferences/sbe_notifications.html",
                             form=form,
                             csrf_token=csrf.field())



# coding=utf-8
"""
"""
from __future__ import absolute_import

from flask import g, url_for as url_for_orig

from abilian.i18n import _l
from abilian.web.action import actions, Action, ModalActionMixin, FAIcon
from abilian.services.security import security, WRITE, MANAGE
from .repository import repository


def url_for(endpoint, **kw):
  return url_for_orig(endpoint, community_id=g.community.slug, **kw)


class CmisContentAction(Action):
  sbe_type = None
  permission = None

  def __init__(self, *args, **kwargs):
    if 'permission' in kwargs:
      self.permission = kwargs.pop('permission')

    Action.__init__(self, *args, **kwargs)

  def pre_condition(self, ctx):
    obj = ctx['object']
    ok = obj.sbe_type == self.sbe_type

    if ok and self.permission is not None:
      ok = self.has_access(self.permission, obj)

    return ok

  def has_access(self, permission, obj):
    return repository.has_permission(g.user, permission, obj)


class BaseFolderAction(CmisContentAction):
  """ Apply to all folders, including root folder
  """
  sbe_type = u'cmis:folder'


class FolderButtonAction(BaseFolderAction):
  """
  """
  _std_template_string = (
    u'<button class="btn {{ action.css_class }}" name="action" '
    u'value="{{ action.name }}" title="{{ action.title }}">'
    u'{{ action.icon }}</button>')

  _modal_template_string = (
    u'<a class="btn {{ action.css_class }}" href="{{ url }}" '
    u'data-toggle="modal" role="button" title="{{ action.title }}">'
    u'{{ action.icon }}</a>')

  def __init__(self, *args, **kwargs):
    self.modal = False

    if 'modal' in kwargs:
      self.modal = kwargs.pop('modal')

    css_class = kwargs.pop('css_class', u'btn-default')
    self.CSS_CLASS = self.CSS_CLASS + u' ' + css_class

    BaseFolderAction.__init__(self, *args, **kwargs)

  @property
  def template_string(self):
    return (self._modal_template_string
            if self.modal
            else self._std_template_string)


class FolderAction(BaseFolderAction):
  """ Apply to all folders except root folder
  """
  sbe_type = u'cmis:folder'

  def pre_condition(self, ctx):
    return (super(FolderAction, self).pre_condition(ctx)
            and ctx['object'] is not repository.root_folder)


class FolderPermisionsAction(BaseFolderAction):
  """ Apply to all folders except root folder
  """
  sbe_type = u'cmis:folder'

  def pre_condition(self, ctx):
    return (super(BaseFolderAction, self).pre_condition(ctx)
            and ctx['object'].depth > 1)


class FolderModalAction(ModalActionMixin, FolderAction):
  """
  """

class DocumentAction(CmisContentAction):
  sbe_type = u'cmis:document'


class DocumentModalAction(ModalActionMixin, DocumentAction):
  """
  """


class RootFolderAction(CmisContentAction):
  """ Apply only for root folder
  """
  def pre_condition(self, ctx):
    return ctx['object'] is repository.root_folder


_checkin_template_action = u'''
<form method="POST" action="{{ url }}" encoding="multipart/form-data" target="_new">
  {{ csrf.field() }}
  <button type="submit" class="btn btn-link" name="action"
          value="{{ action.name}}"
     onclick="window.setTimeout(function () { window.location.reload() }, 500)">
    {%- if action.icon %}{{ action.icon }} {% endif %}
    {{ action.title }}
  </button>
</form>
'''

_lock_template_action = u'''
<form method="POST" action="{{ url }}" encoding="multipart/form-data">
  {{ csrf.field() }}
  <button type="submit" class="btn btn-link" name="action"
          value="{{ action.name}}">
    {%- if action.icon %}{{ action.icon }} {% endif %}
    {{ action.title }}
  </button>
</form>
'''


_actions = (
  # Folder listing action buttons ##########
  FolderButtonAction(
    'documents:folder-listing', 'download', _l(u'Download'),
    icon='download'),
  FolderButtonAction(
    'documents:folder-listing', 'move-files', _l(u'Move to another folder'),
    icon='move', url='#modal-move-files', modal=True,
    permission=WRITE),
  FolderButtonAction(
    'documents:folder-listing', 'delete', _l(u'Delete'),
    permission=WRITE,
    icon='trash', css_class='btn-danger'),
  FolderButtonAction(
    'documents:folder-listing', 'change-owner', _l(u'Change owner'),
    icon='user', url='#modal-change-owner', modal=True,
    permission=MANAGE,
  ),
  # Folder left bar actions ##########
  # view
  RootFolderAction(
    'documents:content', 'view', _l(u'List content'), icon='list',
    condition=(lambda ctx: security.has_role(g.user, "admin")),
    url=lambda ctx: url_for(".folder_view", folder_id=ctx['object'].id),),
  # view
  FolderAction(
    'documents:content', 'view', _l(u'List content'), icon='list',
    url=lambda ctx: url_for(".folder_view", folder_id=ctx['object'].id),),
  # Descendants
  FolderAction(
    'documents:content', 'descendants', _l(u'View descendants'),
    icon=FAIcon('code-fork fa-rotate-90'),
    url=lambda ctx: url_for(".descendants_view", folder_id=ctx['object'].id),),
  # upload
  FolderModalAction(
    'documents:content', 'upload_files', _l('Upload file(s)'),
    icon='upload', url='#modal-upload-files',
    permission=WRITE),
  # edit
  FolderModalAction(
    'documents:content', 'edit', _l('Edit properties'),
    icon='pencil', url='#modal-edit',
    permission=WRITE),
  # new folder
  FolderModalAction(
    'documents:content', 'new_folder', _l('New folder'),
    icon='plus', url='#modal-new-folder',
    permission=WRITE),

  # # members
  # FolderAction(
  #   'documents:content', 'members', _l('Members'), icon='user',
  #   url=lambda ctx: url_for(".members", folder_id=ctx['object'].id)),

  # permissions
  FolderPermisionsAction(
    'documents:content', 'permissions', _l('Permissions'), icon='lock',
    url=lambda ctx: url_for(".permissions", folder_id=ctx['object'].id),
    permission=MANAGE),

  # Document actions ##########
  # View / preview in browser
  DocumentAction(
      'documents:content', 'preview', _l(u'View in browser'),
      icon='eye-open',
      url=lambda ctx: url_for('.document_preview', doc_id=ctx['object'].id),
      condition=(
          lambda ctx:
          ctx['object'].antivirus_ok
          and ctx['object'].content_type in ('text/html',
                                             'text/plain',
                                             'application/pdf'))
  ),
  # edit
  DocumentModalAction(
    'documents:content', 'edit', _l(u'Edit properties'),
    icon='pencil', url='#modal-edit',
    permission=WRITE),
  # Checkin / Checkout
  DocumentAction(
    'documents:content', 'checkout', _l(u'Checkout (Download for edit)'),
    icon='download',
    url=lambda ctx: url_for('.checkin_checkout', doc_id=ctx['object'].id),
    condition=lambda ctx: ctx['object'].lock is None,
    template_string=_checkin_template_action,
  ),
  # DocumentAction(
  #   'documents:content', 'lock', _l(u'Lock for edit'),
  #   icon='lock',
  #   url=lambda ctx: url_for('.checkin_checkout', doc_id=ctx['object'].id),
  #   condition=lambda ctx: ctx['object'].lock is None,
  #   template_string=_lock_template_action,
  # ),
  DocumentAction(
    'documents:content', 'unlock', _l(u'Unlock'),
    icon=FAIcon('unlock'),
    url=lambda ctx: url_for('.checkin_checkout', doc_id=ctx['object'].id),
    condition=lambda ctx: ctx['object'].lock is not None,
    template_string=_lock_template_action,
  ),
  # upload-new / checkin
  DocumentModalAction(
    'documents:content', 'upload', _l(u'Upload new version'),
    icon='upload', url='#modal-upload-new-version',
    # either not locked, either user is owner
    condition=lambda ctx: not ctx['object'].lock or ctx['object'].lock.is_owner(),
    permission=WRITE),
  # send by email
  DocumentModalAction(
      'documents:content', 'send_by_email', _l(u'Send by email'),
      icon='envelope', url='#modal-send-by-email',
      condition=lambda ctx: ctx['object'].antivirus_ok,
      permission=WRITE),
  # delete
  DocumentModalAction(
      'documents:content', 'delete', _l(u'Delete'),
      icon='trash', url='#modal-delete',
      permission=WRITE),
  # refresh preview
  DocumentAction(
      'documents:content', 'refresh_preview', _l(u'Refresh preview'),
      icon='refresh',
      url=lambda ctx: url_for('.refresh_preview', doc_id=ctx['object'].id),
      condition=lambda ctx: ctx['object'].antivirus_ok,
      permission=MANAGE,),
  )


def register_actions(state):
  if not actions.installed(state.app):
    return

  with state.app.app_context():
    actions.register(*_actions)



# coding=utf-8
"""
"""
from __future__ import absolute_import

from flask import render_template

from abilian.core.models.subjects import User
from abilian.web import url_for

from abilian.sbe.testing import BaseTestCase
from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase
from abilian.sbe.apps.communities.models import WRITER
from abilian.sbe.apps.notifications.tasks.social import (
     CommunityDigest, generate_unsubscribe_token,
)

class TestNotificationViews(BaseTestCase):

  def get_setup_config(self):
    cfg = super(TestNotificationViews, self).get_setup_config()
    # "unsubscribe" url must be accessible without csrf, so user can post even
    # if he's not authentified
    cfg.CSRF_ENABLED = True
    cfg.WTF_CSRF_ENABLED = True
    cfg.NO_LOGIN = False
    return cfg

  def setUp(self):
    super(TestNotificationViews, self).setUp()
    self.user = User(email=u'user_1@example.com', password='azerty',
                     can_login=True)
    self.session.add(self.user)
    self.session.commit()

  def test_unsubscribe(self):
    preferences = self.app.services['preferences']
    preferences.set_preferences(self.user, **{'sbe:notifications:daily': True})
    token = generate_unsubscribe_token(self.user)
    url = url_for('notifications.unsubscribe_sbe', token=token)

    r = self.client.get(url)
    self.assert_200(r)
    prefs = preferences.get_preferences(self.user)
    self.assertTrue(prefs['sbe:notifications:daily'])

    r = self.client.post(url)
    self.assert_200(r)
    prefs = preferences.get_preferences(self.user)
    self.assertFalse(prefs['sbe:notifications:daily'])


class NotificationTestCase(CommunityBaseTestCase):

  def setUp(self):
    super(NotificationTestCase, self).setUp()
    self.user = User(email=u'user_1@example.com', password='azerty',
                     can_login=True)
    self.session.add(self.user)
    self.community.set_membership(self.user, WRITER)
    self.session.commit()


  def test_mail_templates(self):
    # this actually tests that templates are parsed without errors, not the
    # rendered content
    digests = [CommunityDigest(self.community)]
    token = generate_unsubscribe_token(self.user)
    render_template("notifications/daily-social-digest.txt",
                    digests=digests, token=token)
    render_template("notifications/daily-social-digest.html",
                    digests=digests, token=token)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import logging

import bleach
from jinja2 import Template, Markup
from flask import render_template_string

from abilian.i18n import _, _l
from abilian.core.util import BasePresenter
from abilian.web.util import url_for

from abilian.sbe.apps.communities.models import Community
from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.documents.views.util import breadcrumbs_for as doc_breadcrumb
from abilian.sbe.apps.forum.models import Post, Thread
from abilian.sbe.apps.wiki.models import WikiPage


logger = logging.getLogger(__name__)

# Poor man's pattern matching.
MESSAGES = {
  ('post', Thread): _l(u'has started conversation "{object}"'),
  ('post', Post): _l(u'has participated in conversation "{object}"'),
  ('post', Document): _l(u'has published document "{object}"'),
  ('post', WikiPage): _l(u'has created wiki page "{object}"'),

  ('update', Document): _l(u'has updated document "{object}"'),
  ('update', WikiPage): _l(u'has updated wiki page "{object}"'),
  ('update', Community): _l(u'has updated community "{object}"'),

  ('join', Community): _l(u'has joined the community {object}.'),
  ('leave', Community): _l(u'has left the community {object}.'),
}

OBJ_TEMPLATE = Template(
  u'<a href="{{ object_url|safe }}">{{ object_name }}</a>'
)

POST_BODY_TEMPLATE = u'''
  <span class="body">"<a href="{{ object_url |safe }}">{{ body }}</a>"
  {%- if post.attachments %}
  <div id="attachments">
    <ul>
      {%- for attachment in post.attachments %}
      <li>
        <span class="attachment-item">
          <img src="{{ attachment.icon }}""
              alt=""/>
          <a href="{{ url_for(attachment) }}">{{ attachment.name }}</a>
          ({{ attachment.content_length|filesize }})
        </span>
      </li>
      {%- endfor %}
    </ul>
  </div>
  {%- endif %}
  </span>
 '''

DOCUMENT_BODY_TEMPLATE = u'''
<div class="body">
  <div>
    <img src="{{ obj.icon }}" style="vertical-align: middle;" alt=""/>
    {% for p in parents[:-1] %}
    <a href="{{ p.path }}">{{ p.label }}</a> <span class="divider">/</span>
    {% endfor %}

    <a href="{{ url_for(obj) }}">{{ obj.name }}</a>
  </div>
  <div>
  {%- if obj.antivirus_ok %}
    <a href="{{ url_for('documents.document_download',
                         community_id=obj.community.slug,
                         doc_id=obj.id,
                         attach=True) }}">
      <i class="glyphicon glyphicon-download"></i>
      {{ _('Download') }} (<small>{{ obj.content_length|filesize  }}</small>)
    </a>
  {%- endif %}
  </div>
</div>
'''


class ActivityEntryPresenter(BasePresenter):
  @property
  def object_url(self):
    return url_for(self._model.object)

  def message(self, ignore_community=False):
    try:
      # another quick&dirty approach for now. FIXME later.
      entry = self._model
      object_class = entry.object_type.split('.')[-1]
      object_class_localized = _(object_class)

      ctx = {}
      ctx['verb'] = entry.verb

      ctx['object_name'] = entry.object.name
      ctx['object_url'] = url_for(entry.object)
      ctx['object_type'] = object_class_localized
      ctx['object'] = OBJ_TEMPLATE.render(**ctx)

      if entry.target:
        ctx['target_name'] = entry.target.name
        ctx['target_url'] = url_for(entry.target)
        ctx['target'] = OBJ_TEMPLATE.render(
          object_name=ctx['target_name'],
          object_url=ctx['target_url']
        )

      msg = MESSAGES.get((entry.verb, entry.object.__class__))
      if msg:
        msg = msg.format(**ctx)
        if entry.target and not ignore_community:
          msg += u" " + _(u'in the community {target}.').format(**ctx)
        else:
          msg += u"."

      elif entry.verb == 'post':
        msg = _(u'has posted an object of type {object_type} '
                u'called "{object}"').format(**ctx)

        if entry.target and not ignore_community:
          msg += u" " + _(u'in the community {target}.').format(**ctx)
        else:
          msg += u"."

      elif entry.verb == 'join':
        msg = _(u'has joined the community {object}.').format(**ctx)

      elif entry.verb == 'leave':
        msg = _(u'has left the community {object}.').format(**ctx)

      elif entry.verb == 'update':
        msg = _(u'has updated {object_type} {object}.').format(**ctx)

      else:
        msg = _(u'has done action "{verb}" on object "{object}".').format(**ctx)

      return Markup(msg)

    except:
      logger.exception('Exception while presenting activity message')
      raise

  @property
  def body(self):
    if isinstance(self.object, Thread):
      body = bleach.clean(self.object.posts[0].body_html, tags=[], strip=True)
      body = Markup(body).unescape()
      if len(body) > 400:
        body = body[0:400] + u"…"
      body = render_template_string(POST_BODY_TEMPLATE,
                                    object_url=self.object_url, body=body, post=self.object.posts[0])
      return Markup(body)
    elif isinstance(self.object, Post):
      body = bleach.clean(self.object.body_html, tags=[], strip=True)
      body = Markup(body).unescape()
      if len(body) > 400:
        body = body[0:400] + u"…"
      body = render_template_string(POST_BODY_TEMPLATE,
                                    object_url=self.object_url, body=body, post=self.object)
      return Markup(body)
    elif isinstance(self.object, Document):
      parents = doc_breadcrumb(self.object)
      body = render_template_string(
        DOCUMENT_BODY_TEMPLATE,
        obj=self.object,
        parents=parents
      )
      return Markup(body)
    else:
      return ""



# coding=utf-8

from flask import g, url_for, current_app
from flask_login import current_user

from abilian.i18n import _l
from abilian.services.security import Admin
from abilian.web.action import actions, Action, ModalActionMixin, FAIcon

class ForumAction(Action):
  def url(self, context=None):
    if self._url or self.endpoint:
      return super(ForumAction, self).url(context=context)

    return url_for("." + self.name, community_id=g.community.slug)


class ThreadAction(ForumAction):
  def pre_condition(self, context):
    thread = actions.context.get('object')
    return not not thread


def is_admin(context):
  svc = current_app.services['security']
  return svc.has_role(current_user, Admin, object=context.get('object'))

def is_in_thread(context):
  thread = context.get('object')
  return not thread


def is_closed(context):
  thread = context.get('object')
  return thread.closed


def not_closed(context):
  return not is_closed(context)


class ForumModalAction(ModalActionMixin, ThreadAction):
  pass


_close_template_action = u'''
<form method="POST" action="{{ url }}" encoding="multipart/form-data">
  {{ csrf.field() }}
  <button type="submit" class="btn btn-link" name="action"
          value="{{ action.name}}">
    {%- if action.icon %}{{ action.icon }} {% endif %}
    {{ action.title }}
  </button>
</form>
'''

_actions = (
  ForumAction('forum:global', 'new_thread',
              _l(u'Start a new conversation'), icon='plus'
  ),
  ForumAction('forum:global', 'index',
              _l(u'Recent conversations'), icon='list'
  ),
  ForumAction('forum:global', 'archives', _l(u'Archives'), icon='calendar'),
  ForumAction('forum:global', 'attachments',
              _l(u'Attachments'), icon='file',
              condition=is_in_thread
  ),
  ForumModalAction('forum:thread', 'delete', _l(u'Delete'),
                   condition=lambda ctx: is_admin(ctx) and not_closed(ctx),
                   url='#modal-delete', icon='trash'
  ),
  ThreadAction('forum:thread', 'close', _l('Close thread'),
               url='close',
               template_string=_close_template_action,
               condition=lambda ctx: is_admin(ctx) and not_closed(ctx),
               icon=FAIcon('lock')
  ),
  ThreadAction('forum:thread', 'reopen', _l('Re-open thread'),
               url='close',
               template_string=_close_template_action,
               condition=lambda ctx: is_admin(ctx) and is_closed(ctx),
               icon=FAIcon('unlock')
  ),
  ThreadAction('forum:thread', 'attachments',
               _l(u'Attachments'), url='attachments', icon='file'
  ),
)


def register_actions(state):
  if not actions.installed(state.app):
    return
  with state.app.app_context():
    actions.register(*_actions)



# coding=utf-8
"""
"""
from __future__ import absolute_import

from urllib import quote
from datetime import datetime

import sqlalchemy as sa
from werkzeug.exceptions import NotFound, BadRequest
from flask import (redirect, request, make_response, flash, g,
                   current_app, render_template)
from flask_mail import Message
from flask_login import current_user

from abilian.i18n import _, render_template_i18n
from abilian.core.extensions import db, mail
from abilian.core.signals import activity
from abilian.services import audit_service
from abilian.services.image import resize, FIT
from abilian.services.conversion import converter
from abilian.web.action import actions
from abilian.web.frontend import add_to_recent_items
from abilian.web.views import default_view
from abilian.web import csrf, url_for
from abilian.sbe.apps.communities.views import default_view_kw

from ..tasks import preview_document, convert_document_content
from ..repository import repository
from ..models import Document
from .util import get_document, check_read_access, breadcrumbs_for, \
  check_write_access, edit_object, match, check_manage_access
from .views import documents


route = documents.route

MAX_PREVIEW_SIZE = 1000


@default_view(documents, Document, id_attr='doc_id', kw_func=default_view_kw)
@route("/doc/<int:doc_id>")
def document_view(doc_id):
  doc = get_document(doc_id)
  check_read_access(doc)
  doc.ensure_antivirus_scheduled()
  db.session.commit()

  bc = breadcrumbs_for(doc)
  actions.context['object'] = doc

  if doc.content_type.startswith("image/"):
    add_to_recent_items(doc, "image")
  else:
    add_to_recent_items(doc, "document")

  has_preview = doc.has_preview()
  audit_entries = audit_service.entries_for(doc)

  ctx = dict(doc=doc, audit_entries=audit_entries, breadcrumbs=bc,
             folder=doc.parent,
             has_preview=has_preview,
             csrf_token=csrf.field())
  return render_template("documents/document.html", **ctx)


#
# Actions on documents
#
@route("/doc/<int:doc_id>", methods=['POST'])
@csrf.protect
def document_edit(doc_id):
  doc = get_document(doc_id)
  check_write_access(doc)

  changed = edit_object(doc)

  if changed:
    db.session.commit()
    flash(_(u"Document properties successfully edited."), "success")
  else:
    flash(_(u"You didn't change any property."), "success")

  return redirect(url_for(doc))


@route("/doc/<int:doc_id>/delete", methods=['POST'])
@csrf.protect
def document_delete(doc_id):
  doc = get_document(doc_id)
  check_write_access(doc)

  parent_folder = doc.parent
  repository.delete_object(doc)
  db.session.commit()

  flash(_(u"File successfully deleted."), "success")
  return redirect(url_for(parent_folder))


@route("/doc/<int:doc_id>/upload", methods=['POST'])
@csrf.protect
def document_upload(doc_id):
  doc = get_document(doc_id)
  check_write_access(doc)

  fd = request.files['file']
  doc.set_content(fd.read(), fd.content_type)
  del doc.lock

  self = current_app._get_current_object()
  activity.send(self, actor=g.user, verb="update", object=doc)
  db.session.commit()
  flash(_(u"New version successfully uploaded"), "success")
  return redirect(url_for(doc))


@route("/doc/<int:doc_id>/download")
def document_download(doc_id, attach=None):
  """Download the file content."""
  doc = get_document(doc_id)

  response = make_response(doc.content)
  response.headers['content-length'] = doc.content_length
  response.headers['content-type'] = doc.content_type

  if attach is None:
    attach = request.args.get('attach')

  if(attach or
     not match(doc.content_type, ("text/plain", "application/pdf", "image/*"))):
    # Note: we omit text/html for security reasons.
    quoted_filename = quote(doc.title.encode('utf8'))
    response.headers['content-disposition'] \
      = 'attachment;filename="{}"'.format(quoted_filename)

  return response


@route('/doc/<int:doc_id>/checkin_checkout', methods=['POST'])
def checkin_checkout(doc_id):
  doc = get_document(doc_id)
  action = request.form.get('action')

  if action not in (u'checkout', u'lock', u'unlock'):
    raise BadRequest(u'Unknown action: %r' % action)

  session = sa.orm.object_session(doc)

  if action in (u'lock', u'checkout'):
    doc.lock = current_user
    d = doc.updated_at
    # prevent change of last modification date
    doc.updated_at = datetime.utcnow()
    session.flush()
    doc.updated_at = d
    session.commit()

    if action == u'lock':
      return redirect(url_for(doc))
    elif action == u'checkout':
      return document_download(doc_id, attach=True)

  if action == u'unlock':
    del doc.lock
    d = doc.updated_at
    # prevent change of last modification date
    doc.updated_at = datetime.utcnow()
    session.flush()
    doc.updated_at = d
    session.commit()
    return redirect(url_for(doc))


def preview_missing_image():
  response = redirect(
    url_for('abilian_sbe_static', filename='images/preview_missing.png'))
  response.headers['Cache-Control'] = 'no-cache'
  return response


@route("/doc/<int:doc_id>/preview_image")
def document_preview_image(doc_id):
  """Returns a preview (image) for the file given by its id."""

  doc = get_document(doc_id)

  if not doc.antivirus_ok:
    return preview_missing_image()

  size = int(request.args.get("size", 0))

  # Just in case
  if size > MAX_PREVIEW_SIZE:
    size = MAX_PREVIEW_SIZE

  # compute image if size != standard document size
  get_image = (converter.get_image
               if size == doc.preview_size
               else converter.to_image)

  content_type = "image/jpeg"

  if doc.content_type.startswith('image/svg'):
    image = doc.content
    content_type = doc.content_type
  elif doc.content_type.startswith("image/"):
    image = doc.content
    if size:
      image = resize(image, size, size, mode=FIT)
  else:
    page = int(request.args.get("page", 0))
    try:
      image = get_image(doc.digest, doc.content, doc.content_type, page, size)
    except:
      # TODO: use generic "conversion failed" image
      image = ""

  if not image:
    return preview_missing_image()

  response = make_response(image)
  response.headers['content-type'] = content_type
  return response


@route("/doc/<int:doc_id>/refresh_preview")
def refresh_preview(doc_id):
  """ Force to compute a new preview
  """
  doc = get_document(doc_id)
  if not doc:
    raise NotFound()

  ct = doc.find_content_type(doc.content_type)
  if ct != doc.content_type:
    doc.content_type = ct
    db.session.commit()

  check_manage_access(doc)
  convert_document_content.apply([doc_id])
  preview_document.apply([doc_id])
  return redirect(url_for(doc))


@route("/doc/<int:doc_id>/send", methods=['POST'])
@csrf.protect
def document_send(doc_id):
  doc = get_document(doc_id)

  recipient = request.form.get("recipient")
  user_msg = request.form.get('message')

  site_name = u'[{}] '.format(current_app.config['SITE_NAME'])
  sender_name = g.user.name
  subject = site_name + _(u'{sender} sent you a file').format(sender=sender_name)
  msg = Message(subject)
  msg.sender = g.user.email
  msg.recipients = [recipient]
  msg.body = render_template_i18n('documents/mail_file_sent.txt',
                                  sender_name=sender_name,
                                  message=user_msg,
                                  document_url=url_for(doc),
                                  filename=doc.title)

  filename = doc.title
  msg.attach(filename, doc.content_type, doc.content)

  mail.send(msg)
  flash(_(u"Email successfully sent"), "success")

  return redirect(url_for(doc))


@route("/doc/<int:doc_id>/preview")
def document_preview(doc_id):
  doc = get_document(doc_id)
  if not doc.antivirus_ok:
    return "Waiting for antivirus to finish"

  if doc.content_type == "application/pdf":
    return redirect(url_for(".document_view_pdf", community_id=g.community.slug, doc_id=doc.id))

  else:
    return redirect(url_for(".document_download", community_id=g.community.slug, doc_id=doc.id))


@route("/doc/<int:doc_id>/view_pdf")
def document_view_pdf(doc_id):
  doc = get_document(doc_id)
  if not doc.antivirus_ok:
    return "Waiting for antivirus to finish"

  return render_template("documents/view_pdf.html",
                         pdf_url=url_for(".document_download", community_id=g.community.slug, doc_id=doc.id))

#
# Tagging (currently not used!)
#
#@route("/tag")
#def tag():
#  tag = request.args.get("tag")
#  if not tag:
#    return redirect("/dm/")
#
#  bc = [dict(path="/", label="Home"), dict(path="/dm/", label="DM")]
#  bc += [dict(path=request.path, label="Filter by tag")]
#  # TODO ...
#  docs = Document.query.filter(Document.tags.like("%" + tag + "%"))
#  docs = list(docs.all())
#  docs = [f for f in docs if tag in f.tags.split(",")]
#  title = "Files filtered by tag: %s" % tag
#  return render_template("dm/home.html", title=title, breadcrumbs=bc,
# files=docs)
#
#
#@route("/<int:file_id>/tag", methods=['POST'])
#def tag_post(file_id):
#  doc = get_document(file_id)
#  tags = request.form.get("tags")
#
#  doc.tags = tags
#  self = current_app._get_current_object()
#  activity.send(self, actor=g.user, verb="tag", object=doc)
#
#  db.session.commit()
#
#  flash("Tags successfully successfully updated", "success")
#  return redirect(url_for(".document_view", doc_id=doc.id))



# coding=utf-8
"""
"""
from __future__ import absolute_import
from future.utils import raise_from

from datetime import datetime, timedelta
import dateutil.parser

from flask import current_app
from flask_login import current_user

from abilian.core.util import utcnow

DEFAULT_LIFETIME = 3600

class Lock(object):
  """
  Represent a lock on a document
  """
  def __init__(self, user_id, user, date, *args, **kwargs):
    self.user_id = user_id
    self.user = user
    if not isinstance(date, datetime):
      try:
        date = dateutil.parser.parse(date)
      except Exception as e:
        raise_from(ValueError('Error parsing date: {!r}'.format(date)), e)

    self.date = date

  @staticmethod
  def new():
    return Lock(current_user.id, unicode(current_user), utcnow())

  def as_dict(self):
    """
    return a dict suitable for serialization to JSON
    """
    return dict(user_id=self.user_id, user=self.user, date=self.date.isoformat())

  @staticmethod
  def from_dict(d):
    """
    Deserialize from a `dict` created by :meth:`as_dict`.
    """
    return Lock(**d)

  @property
  def lifetime(self):
    return current_app.config.get('SBE_LOCK_LIFETIME', DEFAULT_LIFETIME)

  @property
  def expired(self):
    return (utcnow() - self.date) > timedelta(seconds=self.lifetime)

  def is_owner(self, user=None):
    if user is None:
      user = current_user

    return self.user_id == user.id



from __future__ import print_function

from flask import make_response, request, Blueprint, render_template, \
    Response
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import Unauthorized, NotFound

from abilian.core.extensions import db
from abilian.sbe.apps.documents.repository import repository

from .parser import Entry
from .renderer import to_xml, Feed

#
# Constants
#


ROOT = "http://localhost:5000/cmis/atompub"

# MIME TYPES (cf. section 3.3 of the CMIS specs)
MIME_TYPE_ATOM_FEED = "application/atom+xml;type=feed"
MIME_TYPE_ATOM_ENTRY = "application/atom+xml;type=entry"
MIME_TYPE_ATOM_SERVICE = "application/atomsvc+xml"
MIME_TYPE_CMIS_ATOM = "application/cmisatom+xml"
MIME_TYPE_CMIS_QUERY = "application/cmisquery+xml"
MIME_TYPE_CMIS_ALLOWABLE_ACTIONS = "application/cmisallowableactions+xml"
MIME_TYPE_CMIS_TREE = "application/cmistree+xml"
MIME_TYPE_CMIS_ACL = "application/cmisacl+xml"

atompub = Blueprint("cmis", __name__, url_prefix="/cmis/atompub")
route = atompub.route


#
# Dummy for now
#
class Logger(object):
  def debug(self, msg):
    print(msg)

log = Logger()


def log_result(result):
  print(78 * "-")
  print("Response:")
  print(result)
  print(78 * "-")


def produces(*mimetypes):
  def decorator(f):
    return f
  return decorator


def consumes(mimetype):
  def decorator(f):
    return f
  return decorator
# NOT WORKING
#  def decorator(f):
#    def g(*args, **kw):
#      assert mimetype == request.content_type
#      return f(*args, **kw)
#    return g
#  return decorator


# NOT WORKING
def render(template, status=200, mimetype=None):
  def decorator(f):
    def g(*args, **kw):
      print(f, args, kw)
      result = f(*args, **kw)
      rendered = render_template(template, **result)
      headers = {'Content-Type': mimetype}
      response = make_response(rendered, status, headers)
      print(response)
      return response
    return g
  return decorator


BOOLEAN_OPTIONS = ['includeAllowableActions', 'includeACL',
                   'includePolicyIds']

def get_options(args):
  d = {}
  for k in BOOLEAN_OPTIONS:
    v = args.get(k, 'false').strip()
    if v == 'true':
      v = True
    elif v in ('false', ''):
      v = False
    else:
      raise Exception("Unexpected parameter value for %s: %s" % (k, v))
    d[k] = v
  return d


def get_document(id):
  doc = repository.get_document_by_id(id)
  if not doc:
    raise NotFound
  return doc


def get_folder(id):
  obj = repository.get_folder_by_id(id)
  if not obj:
    raise NotFound
  return obj


def get_object(id):
  obj = repository.get_object_by_id(id)
  if not obj:
    raise NotFound
  return obj


#
# Authentication (basic)
#
#@atompub.before_request
def authenticate():
  if not request.authorization:
    raise Unauthorized()

  username = request.authorization.username
  password = request.authorization.password

  if (username, password) != ('admin', 'admin'):
    raise Unauthorized()


#@atompub.errorhandler(401)
def custom_401(error):
  print("custom_401")
  return Response('Authentication required', 401,
                  { 'WWWAuthenticate': 'Basic realm="Login Required"'})


@atompub.errorhandler(NoResultFound)
def not_found_error_handler(error):
  """Converts SQLAlchemy NoResultFound exception to an HTTP error."""
  return "This object does not exist", 404


#
# Service Document
#
@route("/")
@produces(MIME_TYPE_ATOM_SERVICE)
#@render("cmis/service.xml", mimetype=MIME_TYPE_ATOM_SERVICE)
def getRepositoryInfo():
  log.debug("repositoryInfo called")

  root_folder = repository.root_folder
  ctx = {'ROOT': ROOT, 'root_folder': root_folder}

  result = render_template("cmis/service.xml", **ctx)
  response = Response(result, mimetype=MIME_TYPE_ATOM_SERVICE)
  return response

  #return {'root': ROOT}


#
# Service Collections
#
@route("/types")
def getTypeChildren():
  log.debug("getTypeChildren called")

  result = render_template("cmis/types.xml", ROOT=ROOT)
  return Response(result, mimetype=MIME_TYPE_ATOM_FEED)


@route("/types", methods=['POST'])
def createType():
  log.debug("createType called")
  raise NotImplementedError()


#
# Entries (Documents, Folders, Relationships, Policies & Items)
#
@route("/entry")
def getObject():
  id = request.args.get('id')
  path = request.args.get('path')

  log.debug("getObject called on id=%s, path=%s" % (id, path))
  log.debug("URL: " + request.url)

  options = get_options(request.args)
  log.debug("Options: %s" % options)

  if id:
    object = repository.get_object_by_id(id)
  elif path:
    object = repository.get_object_by_path(path)
  else:
    raise NotFound("You must supply either an id or a path.")
  if not object:
    raise NotFound("Object not found")

  result = to_xml(object, **options)
  return Response(result, mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/entry", methods=['PUT'])
def updateProperties():
  id = request.args.get('id')
  if not id:
    path = request.args.get('path')
  else:
    path = ""
  log.debug("updateProperties called on id=%s, path=%s" % (id, path))
  log.debug("URL: " + request.url)

  object = get_object(id)
  return Response(to_xml(object), mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/entry", methods=['DELETE'])
def deleteObject():
  id = request.args.get('id')
  if not id:
    path = request.args.get('path')
  else:
    path = ""
  log.debug("deleteObject called on id=%s, path=%s" % (id, path))
  log.debug("URL: " + request.url)

  object = get_object(id)

  # TODO: remove
  parent = object.parent
  if parent:
    child_count_0 = len(parent.children)

  db.session.delete(object)
  db.session.commit()

  if parent:
    child_count_1 = len(parent.children)
    assert child_count_1 == child_count_0 - 1

  return ("", 204, {})


#
# Content Streams Ressource
#
@route("/content")
def getContentStream():
  id = request.args.get('id')
  log.debug("getContentStream called on " + id)

  document = get_document(id)
  return Response(document.content, mimetype=document.content_type)


# setContentStream + appendContentStream
@route("/content", methods=['PUT'])
def setContentStream():
  id = request.args.get('id')
  log.debug("setContentStream called on " + id)

  document = get_document(id)
  created = document.content is None

  document.content = request.data
  document.content_type = request.content_type
  db.session.commit()

  if created:
    return ("", 201, {})
  else:
    return ("", 204, {})


@route("/content", methods=['DELETE'])
def deleteContentStream():
  id = request.args.get('id')
  log.debug("deleteContentStream called on " + id)

  document = get_document(id)
  document.content = None
  db.session.commit()

  return ("", 204, {})


#
# Allowable Actions Resource
#
@route("/allowableactions")
def getAllowableActions():
  id = request.args.get('id')
  log.debug("getAllowableActions called on " + id)

  object = get_object(id)
  args = {'ROOT': ROOT, 'object': object}
  result = render_template("cmis/allowableactions.xml", **args)
  return Response(result, mimetype=MIME_TYPE_CMIS_ALLOWABLE_ACTIONS)


#
# Type Entries
#
@route("/type")
def getTypeDefinition():
  type_id = request.args.get('id')
  log.debug("getTypeDefinition called on " + type_id)

  if type_id == 'cmis:document':
    result = render_template("cmis/type-document.xml", ROOT=ROOT)
  elif type_id == 'cmis:folder':
    result = render_template("cmis/type-folder.xml", ROOT=ROOT)
  elif type_id == 'cmis:relationship':
    result = render_template("cmis/type-relationship.xml", ROOT=ROOT)
  elif type_id == 'cmis:policy':
    result = render_template("cmis/type-policy.xml", ROOT=ROOT)
  else:
    raise NotImplementedError()

  return Response(result, mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/type", methods=['POST'])
def updateType():
  raise NotImplementedError()


@route("/type", methods=['DELETE'])
def deleteType():
  raise NotImplementedError()

  return ("", 204, {})


#
# Folder Children collection
#
@route("/children")
def getChildren():
  id = request.args.get('id')
  log.debug("getChildren called on " + id)
  log.debug("URL: " + request.url)

  folder = get_folder(id)

  args = {'ROOT': ROOT, 'folder': folder, 'to_xml': to_xml}
  result = render_template("cmis/children.xml", **args)
  log_result(result)
  return Response(result, mimetype=MIME_TYPE_ATOM_ENTRY)


@route("/children", methods=['POST'])
def createObject():
  id = request.args.get('id')
  log.debug("createObject called on " + id)
  log.debug("URL: " + request.url)

  print("Received:")
  print(request.data)

  folder = get_folder(id)

  entry = Entry(request.data)
  name = entry.name
  type = entry.type

  if type == 'cmis:folder':
    new_object = folder.create_subfolder(name)
    db.session.commit()

  elif type == 'cmis:document':
    new_object = folder.create_document(name)
    if entry.content:
      new_object.content = entry.content
      new_object.content_type = entry.content_type
    db.session.commit()

  else:
    raise Exception("Unknown object type: %s" % type)

  result = to_xml(new_object)
  log_result(result)
  return Response(result, status=201, mimetype=MIME_TYPE_ATOM_ENTRY)

  # TODO:
  # URI newloc = null;
  # try {
  #   newloc = new URI(getBase() + "/node/" + newdoc.getId());
  # } catch (URISyntaxException e) {
  # // Shouldn't happen.
  # e.printStackTrace();
  # }

  # String output = getTemplate("entry.ftl").arg("entry", newdoc)
  # .arg("parent", parent).arg("objectType", objectType).render();

  # // XXX: .type() is here because of a bug in resteasy
  # return Response.created(newloc).entity(output).type(MIME_TYPE_ATOM_ENTRY).build();


#
# Feeds
#

# Object Parents Feed (GET)
@route("/parents")
def getObjectParents():
  id = request.args.get('id')
  log.debug("getObjectParents called on " + id)

  object = get_object(id)
  if object.parent:
    feed = Feed(object, [object.parent])
  else:
    feed = Feed(object, [])
  result = feed.to_xml()
  return Response(result, mimetype=MIME_TYPE_ATOM_FEED)


# Changes Feed (GET)
@route("/changes")
def getContentChanges():
  log.debug("getContentChanges called")
  raise NotImplementedError()


# Folder Descendants Feed (GET, DELETE)
@route("/descendants")
def getDescendants():
  id = request.args.get('id')
  log.debug("getObjectParents called on " + id)
  raise NotImplementedError()


@route("/descendants", methods=['DELETE'])
@route("/foldertree", methods=['DELETE'])
def deleteTree():
  id = request.args.get('id')
  log.debug("deleteTree called on " + id)

  object = get_object(id)
  db.session.delete(object)
  db.session.commit()

  return ("", 204, {})


# Folder Tree Feed (GET, DELETE: see above)
@route("/foldertree")
def getFolderTree():
  id = request.args.get('id')
  log.debug("getFolderTree called on " + id)
  raise NotImplementedError()


# All Versions Feed (GET)
def getAllVersions():
  pass


# Type Descendants Feed (GET)
@route("/typedesc")
def getTypeDescendants():
  type_id = request.args.get('typeId')
  log.debug("getTypeDescendants called on " + type_id)

  feed = Feed({}, [])
  result = feed.to_xml()
  return Response(result, mimetype=MIME_TYPE_ATOM_FEED)


#
# Query Collection
#
@route("/query", methods=['POST'])
def query():
  q = request.form.get('q')
  log.debug("query called: " + q)
  raise NotImplementedError()



# coding=utf-8
"""
"""
from __future__ import absolute_import

from flask import render_template, Markup


class UserPhotoInputWidget(object):
  """
  """

  def __call__(self, field, **kwargs):
    kwargs.setdefault('id', field.id)
    user_id = None

    if 'widget_options' in kwargs:
      options = kwargs.pop('widget_options')
      user_id = options.get('user_id')

    return Markup(render_template('social/widget_user_photo.html',
                                  field=field, attrs=kwargs, user_id=user_id))
